//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// FMarkerWin.h - Window for maintaining marker list
//
// $Id: FMarkerWin.h,v 1.9 2001/09/11 10:08:37 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#ifndef FMARKERWIN_H
#define FMARKERWIN_H

#include <qwidget.h>
#include <qpushbutton.h>
#include <qlistview.h>
#include <qlineedit.h>
#include <qvalidator.h>

#include <list>
#include <fstream.h>

#include "FieldObject.h"

class FMarkerWin : public QWidget
{
  Q_OBJECT

public:
  FMarkerWin( const char *filename, QWidget * parent=0, const char * name=0 )
    throw (Exception);
  ~FMarkerWin()
    throw (Exception);

  const FieldMarker *getFirstMarker( int time );
  const FieldMarker *getNextMarker();

public slots:
  void setValues( Vector2d pos, int time, int duration=1 );
  void addMarker( Vector2d pos, int time, int duration, const char *comm );
  void clearList();
  
signals:
  void selectedMarker( int time );
  void closedWindow();

protected slots:
  void gotoButtonClicked();
  void addButtonClicked();
  void resetButtonClicked();
  void closeButtonClicked();
  void currentChanged( QListViewItem *);
  void doubleClicked( QListViewItem *);

protected:
  virtual void closeEvent ( QCloseEvent *e )
  { e->accept(); emit closedWindow(); }
  virtual void resizeEvent( QResizeEvent * );

private:
  QListView   *listView;
  QLineEdit   *timeEdit;
  QLineEdit   *xEdit;
  QLineEdit   *yEdit;
  QLineEdit   *durtnEdit;
  QLineEdit   *commEdit;
  QPushButton *gotoButton;
  QPushButton *addButton;
  QPushButton *resetButton;
  QPushButton *closeButton;

  char        *fname;

  list<FieldMarker *> fmlist;
  // for returning markers to a certain timestep
  list<FieldMarker *>::iterator curMarker;

  int time;  
  
  // get some of the marker properties of a ListViewItem
  static int getTime(QListViewItem *it);
  static int getDurtn(QListViewItem *it);
  static Vector2d getPos(QListViewItem *it);

  // for constructor and destructor
  void readFromFile() throw (Exception);
  void writeToFile() throw (Exception);
};

#endif
