//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
// FieldObject.h - Objects that can be displayed on a field map
// $Id: FieldObject.C,v 1.34 2002/06/12 17:41:11 goetz Exp $
// Time-stamp: <2001-06-26 12:49:32 goetz>
//////////////////////////////////////////////////////////////////////

#include "snapshot.h"
#include "constants.h"

#include "FieldObject.h"
#include "FieldMap.h"
#include "FColors.h"

#include "stdio.h"

#undef DRAWTRACE

//#define DRAWTRACE(output) cout << "FieldObject: "<< output << endl

#ifndef DRAWTRACE
#define DRAWTRACE(output)
#else
#warning: FieldObject debug trace DRAWTRACE is activated!
#endif

// length of the viewcone chord
#define VIEWCONE_LENGTH 10
// length of marker square size
#define MARKER_SIZE 1

static char *num2str[] = {"? ", "1 ", "2 ", "3 ", "4 ", "5 ",
			 "6 ", "7 ", "8 ", "9 ", "10", "11"};

QTNSPC(PenStyle) FieldObject::getStyle(const MStyle style)
{
  switch (style) {
  case MDash:
    return QTNSPC(DashLine);
  case MDot: 
    return QTNSPC(DotLine);
  case MDashDot:
    return QTNSPC(DashDotLine);
  case MDashDotDot:
    return QTNSPC(DashDotDotLine);
  default:
    return QTNSPC(SolidLine);
  }
}

void FieldBackground::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("FieldBackground");

  qp.setBrush( FColors::background );
  qp.drawRect( fm->rect() );

  qp.setPen( FColors::lines );
  qp.setBrush( FColors::field );

  // draw the playing field incluing the lines around it
  qp.drawRect( QRect( fm->getWP( -FIELD_LENGTH/2, -FIELD_WIDTH/2 ),
		      fm->getWP( FIELD_LENGTH/2, FIELD_WIDTH/2 ) ));
  // no filled widgets at the moment!
  qp.setBrush( QTNSPC(NoBrush) );
  
  // center line
  qp.drawLine( fm->getWP( 0, -FIELD_WIDTH/2 ),
	       fm->getWP( 0, FIELD_WIDTH/2 ) );
  // center circle
  qp.drawEllipse( QRect( fm->getWP( -CENTER_CIRCLE_R, -CENTER_CIRCLE_R),
			 fm->getWP( CENTER_CIRCLE_R, CENTER_CIRCLE_R) ));
  // left penalty area
  qp.drawRect( QRect( fm->getWP( -FIELD_LENGTH/2, -PENALTY_AREA_WIDTH/2),
		      fm->getWP( -FIELD_LENGTH/2+PENALTY_AREA_LENGTH,
				 PENALTY_AREA_WIDTH/2 ) ));
  // left goal area
  qp.drawRect( QRect( fm->getWP( -FIELD_LENGTH/2, -GOAL_AREA_WIDTH/2),
		      fm->getWP( -FIELD_LENGTH/2+GOAL_AREA_LENGTH,
				 GOAL_AREA_WIDTH/2) ));
  // right penalty area
  qp.drawRect( QRect( fm->getWP( FIELD_LENGTH/2-PENALTY_AREA_LENGTH,
				 -PENALTY_AREA_WIDTH/2),
		      fm->getWP( FIELD_LENGTH/2,
				 PENALTY_AREA_WIDTH/2 ) ));
  // right goal area
  qp.drawRect( QRect( fm->getWP( FIELD_LENGTH/2-GOAL_AREA_LENGTH, 
				 -GOAL_AREA_WIDTH/2),
		      fm->getWP( FIELD_LENGTH/2,
				 GOAL_AREA_WIDTH/2) ));

  qp.setPen( QPen( FColors::goal ));
  qp.setBrush( QBrush( FColors::goal ));
  
  // left goal
  qp.drawRect( QRect( fm->getWP( -FIELD_LENGTH/2 - 1, -GOAL_WIDTH/2 ),
		      fm->getWP( -FIELD_LENGTH/2, GOAL_WIDTH/2 )) );
  // right goal
  qp.drawRect( QRect( fm->getWP( FIELD_LENGTH/2, -GOAL_WIDTH/2 ),
		      fm->getWP( FIELD_LENGTH/2 + 1, GOAL_WIDTH/2 )) );
}

void FieldBall::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("FieldBall");

  qp.setPen( FColors::ballPen );
  qp.setBrush( FColors::ballFill );
  
  double bsz(BALL_SIZE * fm->ballmag() );

  qp.drawEllipse( QRect( fm->getWP( pos.getX()-bsz,
				    pos.getY()-bsz ),
			 fm->getWP( pos.getX()+bsz,
				    pos.getY()+bsz) ));
  if (fm->drawKickRange()) {
    qp.setBrush( QTNSPC(NoBrush) );
    double krng(BALL_SIZE + PLAYER_SIZE + .8*KICKABLE_MARGIN);
    qp.drawEllipse( QRect( fm->getWP( pos.getX()-krng,
				      pos.getY()-krng ),
			   fm->getWP( pos.getX()+krng,
				      pos.getY()+krng) ));
  }
}

void WFieldBall::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("WFieldBall");

  qp.setPen( color );
  qp.setBrush( QTNSPC(NoBrush) );
  
  double bsz(BALL_SIZE * fm->ballmag() );

  qp.drawEllipse( QRect( fm->getWP( pos.getX()-bsz,
				    pos.getY()-bsz ),
			 fm->getWP( pos.getX()+bsz,
				    pos.getY()+bsz) ));
  // velocity
  qp.drawLine( fm->getWP( pos ), 
	       fm->getWP( pos + vel * fm->speedmag() ) );
  // age
  qp.drawLine( fm->getWP( pos.getX()-bsz,
			  pos.getY()-bsz ) - QPoint(0,2),
	       fm->getWP( pos.getX()-bsz + age*bsz,
			  pos.getY()-bsz ) - QPoint(0,2));
}

void FieldPlayer::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("FieldPlayer");

  // if view cone is wanted, draw it!
  if (fm->showViewCone( leftTeam, num )) {
    // draw three parts separately
    qp.setPen( QPen(QTNSPC(NoPen)));
    qp.setRasterOp(QTNSPC(OrROP));

    qp.setBrush( FColors::viewFeel);
    qp.drawEllipse( QRect(fm->getWP( pos.getX()-VISIBLE_DISTANCE,
				     pos.getY()-VISIBLE_DISTANCE ),
			  fm->getWP( pos.getX()+VISIBLE_DISTANCE,
				     pos.getY()+VISIBLE_DISTANCE)));
    
    qp.setBrush( FColors::viewNear);
    qp.drawPie( QRect(fm->getWP( pos.getX()-UNUM_FAR_LENGTH,
				   pos.getY()-UNUM_FAR_LENGTH ),
			fm->getWP( pos.getX()+UNUM_FAR_LENGTH,
				   pos.getY()+UNUM_FAR_LENGTH )),
		  -int(RAD_TO_DEG(headdir - viewWidth / 2)*16),
		  -int(RAD_TO_DEG(viewWidth)*16));

    qp.setBrush( FColors::viewFar);
    qp.drawPie( QRect(fm->getWP( pos.getX()-TEAM_FAR_LENGTH,
				   pos.getY()-TEAM_FAR_LENGTH ),
			fm->getWP( pos.getX()+TEAM_FAR_LENGTH,
				   pos.getY()+TEAM_FAR_LENGTH )),
		  -int(RAD_TO_DEG(headdir - viewWidth / 2)*16),
		  -int(RAD_TO_DEG(viewWidth)*16));

    qp.setRasterOp(QTNSPC(CopyROP));
  }

  double psz(PLAYER_SIZE * fm->playermag() );

  QRect playerRect= QRect( fm->getWP( pos.getX()-psz,
				      pos.getY()-psz ),
			   fm->getWP( pos.getX()+psz,
				      pos.getY()+psz) );

  if (kicking)
    qp.setPen( FColors::playerKick );
  else
    qp.setPen( FColors::playerNorm );

  qp.setBrush( QTNSPC(NoBrush) );

  qp.drawEllipse( playerRect ); 

  if (isGoalie)
    if (leftTeam)
      qp.setBrush( FColors::leftGoalie );
    else
      qp.setBrush( FColors::rightGoalie );
  else    
    if (leftTeam)
      qp.setBrush( FColors::leftTeam );
    else
      qp.setBrush( FColors::rightTeam );
  
  // body direction
  qp.drawPie( playerRect, -int(RAD_TO_DEG(bodydir - PI_2)*16),-180*16 );

  // "tentacles" to indicate head angle
  qp.drawLine( fm->getWP( pos ), 
	       fm->getWP( pos.getX() + 2 * psz * 
			  cos(headdir + viewWidth/2),
			  pos.getY() + 2 * psz * 
			  sin(headdir + viewWidth/2) ));
  qp.drawLine( fm->getWP( pos ), 
	       fm->getWP( pos.getX() + 2 * psz * 
			  cos(headdir - viewWidth/2),
			  pos.getY() + 2 * psz * 
			  sin(headdir - viewWidth/2) ));
			  
  // player number
  qp.drawText( fm->getWP( pos.getX()-psz,
			  pos.getY()+psz) + QPoint(-2,8),
	       (num < 0) || (num > 11) ? "?" : num2str[num] );
  // if stamina display is desired, do so!
  if ((stamina > -1) || (het_type > -1)) {
    char sbuf[20]= "", tbuf[10]= "";
    if (stamina > -1) 
      sprintf(sbuf, "%d", stamina);
    if (het_type > -1)
      sprintf(tbuf, "T%d", het_type);
    if ((stamina > -1) && (het_type > -1))
      strcat(sbuf, " ");
    strcat(sbuf, tbuf);
    qp.drawText( fm->getWP( pos.getX()-psz, 
			    pos.getY()+psz) + QPoint(-2,20),
		 sbuf);
  }

  // if collision, draw a double pink circle around player
  if (collision) {
    qp.setPen( FColors::playerColl );
    qp.setBrush( QTNSPC(NoBrush) );
    qp.drawEllipse( playerRect );
  }
}

// draw a player as stored in the worldmodel (no special drawing for "me")
void WFieldPlayer::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("WFieldPlayer");

  double psz(PLAYER_SIZE * fm->playermag() );

  QRect playerRect= QRect( fm->getWP( pos.getX()-psz,
				      pos.getY()-psz ),
			   fm->getWP( pos.getX()+psz,
				      pos.getY()+psz));
			   
  qp.setPen( QPen( color ));
  qp.setBrush( QTNSPC(NoBrush) );
  
  // body direction
  qp.drawPie( playerRect, int(RAD_TO_DEG(-bodydir - PI_2)*16), 90*16 );
  qp.drawPie( playerRect, int(RAD_TO_DEG(-bodydir)*16), 90*16 );
  
  // player body
  qp.drawEllipse( playerRect );
  // velocity
  qp.drawLine( fm->getWP( pos ), 
	       fm->getWP( pos + vel * fm->speedmag() ) );
  // Number, team name
  char buf[6] = "";
  if ( leftTeam )
    strcpy( buf, "L");
  else
    strcpy( buf, "R");
  if ( isGoalie )
    strcat( buf, "G " );
  else 
    strcat( buf, " " );
  if ( (num < 0) || (num > 11) )
    strcat( buf, "?" );
  else
    strncat( buf, num2str[num], 2 );

  // player number
  qp.drawText( fm->getWP( pos.getX()-psz,
			  pos.getY()+psz) + QPoint(-2,8),
	       buf );
  // age
  qp.drawLine( fm->getWP( pos.getX()-PLAYER_SIZE,
			  pos.getY()-PLAYER_SIZE ) - QPoint(0,2),
	       fm->getWP( pos.getX()-PLAYER_SIZE + age*PLAYER_SIZE,
			  pos.getY()-PLAYER_SIZE ) - QPoint(0,2));
}

void FieldCircle::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("FieldCircle");

  qp.setPen( QPen( color, 1, getStyle(style) ) );
  qp.setBrush( QTNSPC(NoBrush) );

  // t is the top left point of the ellipse rectangle,
  // b the bottom right one
  QPoint c(fm->getWP(pos)), 
    t( fm->getWP( pos.getX()-rad, pos.getY()-rad ) ),
    b( fm->getWP( pos.getX()+rad, pos.getY()+rad ) );

  qp.drawEllipse( QRect( t, b) );
  
  if ( hasgrid ) {
    qp.drawLine( t.x(), c.y(), b.x(), c.y() );
    qp.drawLine( c.x(), t.y(), c.x(), b.y() );
  }
}

void FieldPosition::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("FieldPosition");

  // a position shall be marked by a circle of radius 1
  // NB: NO "X" is marking the spot !!!
  FieldCircle(pos, PLAYER_SIZE*fm->playermag() , true, color, MSolid).draw( qp, fm );
}
  
void FieldLine::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("FieldLine");

  qp.setPen( QPen( color, 1, getStyle( style ) ) );
  qp.drawLine( fm->getWP( pos ), fm->getWP( endPos ));
}

void FieldVLine::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("FieldVLine");

  FieldLine( pos, pos+ Vector2d( 0,FIELD_WIDTH ), color, style).draw(qp, fm);
}

void FieldBox::draw( QPainter & qp, const FieldMap *fm ) const
{
  DRAWTRACE("FieldBox");

  qp.setBrush( QTNSPC(NoBrush) );
  qp.setPen( QPen( color, 1, getStyle( style ) ) );
  qp.drawRect( QRect(fm->getWP( pos ), fm->getWP( other )));
}

void FieldCorridor::draw( QPainter & qp, const FieldMap *fm ) const
{  
  DRAWTRACE("FieldCorridor");

  Angle ang ( (target - pos).getAngle() ),
    angp ( (target - pos).perp().getAngle() ),
    angppp ( (target - pos).perp().perp().perp().getAngle() );

  qp.setPen( QPen( color, 1, getStyle(style)) );
  
  qp.moveTo( fm->getWP( pos ) );
  qp.lineTo( fm->getWP( pos + 
			radAng2Vector2d( 0.5 * diam, angp ) +
			radAng2Vector2d( 0.5 * diam / tan( alpha ), ang )) );
  qp.lineTo( fm->getWP( target + radAng2Vector2d( 0.5 * diam, angp )) );
  qp.lineTo( fm->getWP( target - radAng2Vector2d( 0.5 * diam, angp )) );
  qp.lineTo( fm->getWP( pos + 
			radAng2Vector2d( 0.5 * diam / tan( alpha ), ang ) +
			radAng2Vector2d( 0.5 * diam, angppp )) );
  qp.lineTo( fm->getWP( pos ) );
}

void FieldLabel::draw( QPainter & qp, const FieldMap *fm ) const
{
  qp.setPen( QPen( color ) );
  qp.drawText( fm->getWP(pos), text.c_str() ); 
}

FieldMarker::FieldMarker( const Vector2d &pos, const int time, 
			  const int duration, const char *comment)
  : FieldObject( pos ), time(time), duration(duration)
{
  const char *buf= comment;
  if (buf == 0)
    buf= "";
  
  comm= new char[strlen(buf)+1];
  // simple way of ignoring double quotes (and everything following)
  sscanf(buf, "%[^\"]", comm);
}


void FieldMarker::draw( QPainter & qp, const FieldMap *fm ) const
{
  QPoint topLeft(fm->getWP(pos.getX() - MARKER_SIZE/2.0,
			   pos.getY() - MARKER_SIZE/2.0)),
    bottomRight(fm->getWP(pos.getX() + MARKER_SIZE/2.0,
			  pos.getY() + MARKER_SIZE/2.0));
  qp.setPen( FColors::marker );
  qp.setBrush( QTNSPC(NoBrush) );

  qp.drawRect( QRect(topLeft, bottomRight) );
  qp.drawLine( topLeft, bottomRight );
  qp.drawLine( topLeft.x(), bottomRight.y(), 
	       bottomRight.x(), topLeft.y() );
}

int FieldMarker::operator==( const FieldMarker &fm )
{
  return time==fm.time && pos==fm.pos && duration==fm.duration;
}

FieldMarker::operator char * () const
{
  char *enc= new char[128];
  sprintf(enc, "(fmark %d %d (%f, %f) \"%s\")", time, duration, 
	  pos.getX(), pos.getY(), comm);
  return enc;
}

ostream & operator << (ostream &os, const FieldMarker &f)
{
  char *s=  (char *)f;
  os << s;
  delete s;
  return os;
}

void ShowinfoType::draw( QPainter & qp, const FieldMap *fm ) const
{
  if (ball != 0)
    ball->draw(qp, fm);
  for (int i=0; i<11; i++) {
    if (ourTeam[i] != 0) 
      ourTeam[i]->draw( qp, fm);
    if (otherTeam[i] != 0)
      otherTeam[i]->draw( qp, fm);
  }
}

ShowinfoType::ShowinfoType()
  : FieldObject( Vector2d() ), ball(0)
{
  for (int i=0; i<11; i++) {
    ourTeam[i]= 0;
    otherTeam[i]= 0;
  }
}

ShowinfoType::~ShowinfoType()
{
  delete ball;
  for (int i=0; i<11; i++) {
    delete ourTeam[i];
    delete otherTeam[i];
  }
}
 
void FieldSnapshot::draw( QPainter & qp, const FieldMap *fm ) const
{
  for (int i=0; i<11; i++) {
    if (ourTeam[i] != 0) 
      ourTeam[i]->draw( qp, fm);
    if (otherTeam[i] != 0)
      otherTeam[i]->draw( qp, fm);
  }
  if (ball != 0)
    ball->draw(qp, fm);
}

FieldSnapshot::FieldSnapshot()
  : FieldObject( Vector2d() ), ball(0)
{
  for (int i=0; i<11; i++) {
    ourTeam[i]= 0;
    otherTeam[i]= 0;
  }
}

FieldSnapshot::~FieldSnapshot()
{
  delete ball;
  for (int i=0; i<11; i++) {
    delete ourTeam[i];
    delete otherTeam[i];
  }
}
