#ifndef SSERVER_H
#define SSERVER_H
#include "SoccerMessage.h"
#include <coordinates.h>

// import structs from soccerserver
namespace SServer {
#include <types.h>
};

/** return showinfo from soccer_msg or NULL if invalid.
    the returned pointer points into msg and has therefore
    the same lifetime.
*/
SServer::showinfo_t2 *getShowInfo(const SoccerMessage &msg) {
  if (msg.sender != SoccerServer)
    return NULL;

  if (string(msg).length() < sizeof(SServer::dispinfo_t2))
    return NULL;

  SServer::dispinfo_t2 *rinfo2 =
    (SServer::dispinfo_t2 *)(string(msg).data());

  char msgType(ntohs((short)(rinfo2->mode)));
  if (msgType != SHOW_MODE)
    return NULL;
  return &(rinfo2->body.show);
}

/* get diverse elements from showinfo */
Vector2d getBallPos(SServer::showinfo_t2 *show2) {
  return Vector2d((long)ntohl(show2->ball.x)/SHOWINFO_SCALE2,
		  (long)ntohl(show2->ball.y)/SHOWINFO_SCALE2);
}

Vector2d getBallVel(SServer::showinfo_t2 *show2) {
  return Vector2d((long)ntohl(show2->ball.deltax)/SHOWINFO_SCALE2,
		  (long)ntohl(show2->ball.deltay)/SHOWINFO_SCALE2);
}

Vector2d getPlayerPos(SServer::showinfo_t2 *show2, int p) {
  return Vector2d((long)ntohl(show2->pos[p].x)/SHOWINFO_SCALE2,
		  (long)ntohl(show2->pos[p].y)/SHOWINFO_SCALE2);
}

bool playerKicked(SServer::showinfo_t2 *show2, int p) {
  return (ntohs(show2->pos[p].mode) & KICK);
}

#endif
