//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// SoccerMsgParser.h - A parser for messages coming from the soccer
//                    server.
//
// $Id: SoccerMsgParser.h,v 1.7 2001/09/11 10:08:38 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#ifndef SOCCERMSGPARSER_H
#define SOCCERMSGPARSER_H

#include "qcolor.h"

#include "FieldObject.h"
#include "FMarkerWin.h"
#include "FLogWindow.h"
#include "SoccerMessage.h"
#include "mrbtypes.h"
#include "coordinates.h"

#include <list>
#include <string>
#include <strstream>

class MainWidget;

class SoccerMsgParser
{
public:
  SoccerMsgParser( MainWidget *mw, FMarkerWin *fmw, FLogWindow *flw );

  FieldObject *parseMessage( const SoccerMessage &msg );
  
  FieldObject *serverMessage( const SoccerMessage &msg );
  
  FieldObject *playerMessage( const SoccerMessage &msg,
			      int plNum, bool leftTeam);
  FieldObject *monitorMessage( const SoccerMessage &msg );

private:
  enum SenderGroup {
    None, Player, Server, FMonitor
  };

  void getSender( const SoccerMessage &msg,
		  SenderGroup &g, int &plNum, bool &leftTeam);

  void parseMarker( int time, const char *msg );

  // side switching functions (for games as right team)
  double sws( const double x, bool leftTeam )
  { return leftTeam ? x : -x; }
  Angle sws( const Angle &a, bool leftTeam )
  { return leftTeam ? a : PI - a; }

  // removes unwanted chars from string
  char *cleanString( const char *s );

  MainWidget *mw;
  FMarkerWin *fmw;
  FLogWindow *flw;
};

#endif
