dnl @synopsis MDL_HAVE_OPENGL
dnl
dnl Search for OpenGL.  We search first for Mesa (a GPL'ed version of
dnl Mesa) before a vendor's version of OpenGL, unless we were
dnl specifically asked not to with `--with-Mesa=no' or `--without-Mesa'.
dnl
dnl The four "standard" OpenGL libraries are searched for: "-lGL",
dnl "-lGLU", "-lGLX" (or "-lMesaGL", "-lMesaGLU" as the case may be) and
dnl "-lglut".
dnl
dnl All of the libraries that are found (since "-lglut" or "-lGLX" might
dnl be missing) are added to the shell output variable "GL_LIBS", along
dnl with any other libraries that are necessary to successfully link an
dnl OpenGL application (e.g. the X11 libraries).  Care has been taken to
dnl make sure that all of the libraries in "GL_LIBS" are listed in the
dnl proper order.
dnl
dnl Additionally, the shell output variable "GL_CFLAGS" is set to any
dnl flags (e.g. "-I" flags) that are necessary to successfully compile
dnl an OpenGL application.
dnl
dnl The following shell variable (which are not output variables) are
dnl also set to either "yes" or "no" (depending on which libraries were
dnl found) to help you determine exactly what was found.
dnl
dnl   have_GL
dnl   have_GLU
dnl   have_GLX
dnl   have_glut
dnl
dnl A complete little toy "Automake `make distcheck'" package of how to
dnl use this macro is available at:
dnl
dnl   ftp://ftp.slac.stanford.edu/users/langston/autoconf/ac_opengl-0.01.tar.gz
dnl
dnl Please note that as the ac_opengl macro and the toy example evolves,
dnl the version number increases, so you may have to adjust the above
dnl URL accordingly.
dnl
dnl @version $Id: acinclude.m4,v 1.8 2002/04/08 15:52:19 cs Exp $
dnl @author Matthew D. Langston <langston@SLAC.Stanford.EDU>

AC_DEFUN(MDL_HAVE_OPENGL,
[
  AC_REQUIRE([AC_PROG_CC])
  AC_REQUIRE([AC_PATH_X])
  AC_REQUIRE([AC_PATH_XTRA])

  AC_CACHE_CHECK([for OpenGL], mdl_cv_have_OpenGL,
  [
dnl Check for Mesa first, unless we were asked not to.
    AC_ARG_ENABLE(Mesa, [--with-Mesa   Prefer the Mesa library over a vendors native OpenGL library (default=yes)], use_Mesa=$enableval, use_Mesa=yes)

    if test x"$use_Mesa" = xyes; then
       GL_search_list="MesaGL   GL"
      GLU_search_list="MesaGLU GLU"
      GLX_search_list="MesaGLX GLX"
    else
       GL_search_list="GL  MesaGL"
      GLU_search_list="GLU MesaGLU"
      GLX_search_list="GLX MesaGLX"
    fi

    AC_LANG_SAVE
    AC_LANG_C

dnl If we are running under X11 then add in the appropriate libraries.
if ! test x"$no_x" = xyes; then
dnl Add everything we need to compile and link X programs to GL_X_CFLAGS
dnl and GL_X_LIBS.
  GL_CFLAGS="$X_CFLAGS"
  GL_X_LIBS="$X_PRE_LIBS $X_LIBS -lX11 -lXext -lXmu -lXt -lXi $X_EXTRA_LIBS"
fi
    GL_save_CPPFLAGS="$CPPFLAGS"
    CPPFLAGS="$GL_CFLAGS"

    GL_save_LIBS="$LIBS"
    LIBS="$GL_X_LIBS"


    # Save the "AC_MSG_RESULT file descriptor" to FD 8.
    exec 8>&AC_FD_MSG

    # Temporarily turn off AC_MSG_RESULT so that the user gets pretty
    # messages.
    exec AC_FD_MSG>/dev/null

    AC_SEARCH_LIBS(glAccum,          $GL_search_list, have_GL=yes,   have_GL=no)
    AC_SEARCH_LIBS(gluBeginCurve,   $GLU_search_list, have_GLU=yes,  have_GLU=no)
    AC_SEARCH_LIBS(glXChooseVisual, $GLX_search_list, have_GLX=yes,  have_GLX=no)
    AC_SEARCH_LIBS(glutInit,        glut,             have_glut=yes, have_glut=no)



    # Restore pretty messages.
    exec AC_FD_MSG>&8

    if test -n "$LIBS"; then
      mdl_cv_have_OpenGL=yes
      GL_LIBS="$LIBS"
      AC_SUBST(GL_CFLAGS)
      AC_SUBST(GL_LIBS)
    else
      mdl_cv_have_OpenGL=no
      GL_CFLAGS=
    fi

dnl Reset GL_X_LIBS regardless, since it was just a temporary variable
dnl and we don't want to be global namespace polluters.
    GL_X_LIBS=

    LIBS="$GL_save_LIBS"
    CPPFLAGS="$GL_save_CPPFLAGS"

    AC_LANG_RESTORE
  ])
])

AC_DEFUN(ADD_CHECK_QT,
[
  AC_REQUIRE([AC_PROG_CC])
  AC_REQUIRE([AC_PATH_X])
  AC_REQUIRE([AC_PATH_XTRA])

  AC_MSG_CHECKING("for Qt")

  AC_ARG_WITH(qt-includes, [--with-qt-includes path to qt headers],
	qt_includes="$withval", qt_includes="$QTDIR/include")
  AC_ARG_WITH(qt-libs, [--with-qt-libs path to qt libs],
	qt_libs="$withval", qt_libs="$QTDIR/lib")
  AC_ARG_WITH(qt-moc,  [--with-qt-moc path to qt moc],
	qt_moc="$withval", qt_moc="$QTDIR/bin/moc")
  dnl Add everything we need to compile and link X programs to QT_X_CFLAGS
  dnl and QT_X_LIBS.
  QT_CFLAGS="$X_CFLAGS"
  if test "$qt_includes" != "" ; then
	QT_CFLAGS="$QT_CFLAGS -I$qt_includes"
  fi
  QT_X_LIBS="$X_PRE_LIBS $X_LIBS -lqt -lX11 -lXext -lXmu $X_EXTRA_LIBS"
  if test "$qt_libs" != "" ; then
	QT_X_LIBS="-L$qt_libs $QT_X_LIBS"
  fi

  QT_save_CPPFLAGS="$CPPFLAGS"
  CPPFLAGS="$QT_CFLAGS"

  QT_save_LIBS="$LIBS"
  LIBS="$QT_X_LIBS"

  dnl check if compiling
  AC_LANG_SAVE
  AC_LANG_CPLUSPLUS
  AC_TRY_COMPILE([
	#include <qapplication.h>
	], [
	QApplication tmp(NULL);
	],
  	[ add_cv_have_Qt=yes; AC_MSG_RESULT("yes"); ],
  	[ AC_MSG_ERROR("Qt is required") ]
  )
  AC_LANG_RESTORE

  if test "$add_cv_have_Qt" == "yes" ; then
	MOC="$qt_moc"
	QT_LIBS="$LIBS"
	QT_INCLUDES="$qt_includes"
	AC_SUBST(QT_CFLAGS)
	AC_SUBST(QT_LIBS)
	AC_SUBST(MOC)
  else
	QT_LIBS=
	QT_CFLAGS=
  fi

  dnl Reset QT_X_LIBS regardless, since it was just a temporary variable
  dnl and we don't want to be global namespace polluters.
  QT_X_LIBS=

  LIBS="$QT_save_LIBS"
  CPPFLAGS="$QT_save_CPPFLAGS"
])

AC_DEFUN(ADD_CHECK_QT_VERSION,
[
  AC_MSG_CHECKING("Qt version");

  _qt_what=$
  _qt_what=${_qt_what}3
	
  _qt_ver=`eval "grep -w '#define QT_VERSION' $QT_INCLUDES/qglobal.h|awk '{print $_qt_what}'"`

  AC_MSG_RESULT($_qt_ver)

  if test $_qt_ver -ge 200 ; then
	CXXFLAGS="$CXXFLAGS -DQT2"
  fi

  _qt_ver=
  _qt_what=
])

AC_DEFUN(ADD_CHECK_SSERVER,
[
  AC_MSG_CHECKING("for soccerserver");

  AC_ARG_WITH(sserver-dir, [--with-sserver-dir path to server dir (rcssserver-*.* rsp. sserver-*.*)],
	sserver_dir="$withval", sserver_dir="/home/robocup/opt/rcsss/latest")

  if test -d $sserver_dir/src ; then
	SSERVER_VERSION=8
	SSERVER_DIR="$sserver_dir/src"
	AC_MSG_RESULT("version 8");
  elif test -d $sserver_dir/server ; then
	SSERVER_VERSION=7
	SSERVER_DIR="$sserver_dir/server"
	AC_MSG_RESULT("version 6/7");
  else
  	AC_MSG_ERROR("soccerserver is required")
  fi

  sserver_save_CPPFLAGS="$CPPFLAGS"
  CPPFLAGS="$CPPFLAGS -I$SSERVER_DIR"

  dnl check if dispinfo_t is found
  AC_LANG_SAVE
  AC_LANG_CPLUSPLUS
  AC_TRY_COMPILE([
	#include <param.h>
	#include <types.h>
	], [
	dispinfo_t test;
	],
  	[ add_cv_have_sserver=yes; ],
  	[ AC_MSG_ERROR("soccerserver source is required") ]
  )
  AC_LANG_RESTORE

  if test "$add_cv_have_sserver" == "yes" ; then
	SSERVER_CFLAGS="-I$SSERVER_DIR"
	if test $SSERVER_VERSION == 7 ; then
		SSERVER_COMMAND="$SSERVER_DIR/soccerserver -sfile funssel_server.conf \
					-pfile $SSERVER_DIR/player.conf"
	else
		SSERVER_COMMAND="$SSERVER_DIR/rcssserver -sfile funssel_server.conf"
	fi
	AC_SUBST(SSERVER_CFLAGS)
	AC_SUBST(SSERVER_DIR)
	AC_SUBST(SSERVER_COMMAND)
	AC_SUBST(SSERVER_VERSION)
  else
	SSERVER_CFLAGS=
  fi

  CPPFLAGS="$sserver_save_CPPFLAGS"
])

