//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// 	$Id: AdvSkills.C,v 1.54 2002/06/03 12:26:38 cm Exp $
//      $Manager: aarn (Axel Arnold)$
//
//////////////////////////////////////////////////////////////////////

// for axels wonderful emacs-extension
// cwz-build-command: (cd ../MasterControl; make all)

#include "constants.h"
#include "worldmodel.h"
#include "AdvSkills.h"
#include "mytrace.h"
#include "minmax.h"
#include <strstream>

// debugging information switches
//#define TRACE_TURNHEAD
//#define TRACE_INTERCEPT
//#define TRACE_DRIBBLE
//#define TRACE_POINTTO
//#define TRACE_TACKLE

#define FUNSSELMACROS
#include "drawMacros.h"

#define TextPos   Vector2d(-50,-30)
#define TextPos1  Vector2d(-50,-25)
#define TextPos2  Vector2d(-50,-20)
#define TextPos3  Vector2d(-50,-15)
#define TextPosA  Vector2d(-50,-10)
#define TextPosA1 Vector2d(-50, -5)
#define TextPosA2 Vector2d(-50,  0)
#define TextPosA3 Vector2d(-50,  5)
#define TextPosA4 Vector2d(-50, 10)

#ifndef  TRACE_INTERCEPT
#define  ICDRAWTRACE(cmd)
#else
#define  ICDRAWTRACE(cmd) Comm::sendMCmd(cmd)
#warning ICDRAWTRACE enabled
#endif

#ifndef  TRACE_TURNHEAD
#define  THDRAWTRACE(cmd)
#else
#define  THDRAWTRACE(cmd) Comm::sendMCmd(cmd)
#warning THDRAWTRACE enabled
#endif

#ifdef TRACE_TURNHEAD
#define THNAMEDLINE(name, start, dir) { \
    Vector2d end(start + dir); \
    THDRAWTRACE(LINE(start, end)); \
    THDRAWTRACE(TEXT(name, end)); }
#else
#define THNAMEDLINE(name, start, dir)
#endif

#ifdef TRACE_INTERCEPT
#define ICNAMEDLINE(name, start, dir) { \
    Vector2d end(start + dir); \
    ICDRAWTRACE(LINE(start, end)); \
    ICDRAWTRACE(TEXT(name, end)); }
#else
#define ICNAMEDLINE(name, start, dir)
#endif

#ifndef  TRACE_DRIBBLE
#define  DRBTRACE(cmd)
#else
#define  DRBTRACE(cmd) Comm::sendMCmd(cmd)
#warning DRBTRACE enabled
#endif

#ifndef  TRACE_POINTTO
#define  PTTRACE(cmd)
#else
#define  PTTRACE(cmd) Comm::sendMCmd(cmd)
#warning PTTRACE enabled
#endif

#ifndef  TRACE_TACKLE
#define  TACDRWTRACE(cmd)
#define  TACTRACE(cmd)
#else
#define  TACDRWTRACE(cmd) Comm::sendMCmd(cmd)
#define  TACTRACE(cmd) cout << cmd << endl;
#warning TACTRACE enabled
#endif

/////////////////////////////////////////////////// ENTIRE SKILLS STUFF

AdvSkills *skills = 0;

/*********************************** The constructor ************************/

AdvSkills::AdvSkills(Comm *c):
  // pseudo constants
  precision(1e-5),
  kickRange(PLAYER_SIZE + BALL_SIZE + 0.8*KICKABLE_MARGIN),
  minSpeed(0.1),
  prepareOuterLimit(3*kickRange),
  prepareInnerLimit(2*kickRange),
  minInterceptTime(30),
  goToThreshStretch(0.008),
  turnMinAngle(DEG_TO_RAD(15)),
  directThreshold(0.1),
  laybackErr(0.25),
  innerSafeCircle(PLAYER_SIZE+BALL_SIZE+0.3*KICKABLE_MARGIN),
  safeCircle(PLAYER_SIZE+BALL_SIZE+0.55*KICKABLE_MARGIN),
  stopableV(0.5*KICK_POWER_RATE*MAXPOWER),
  dribbleBodyErr(DEG_TO_RAD(10)),
  dribbleHoldErr(DEG_TO_RAD(30)),
  dribbleMaxV(PLAYER_SPEED_MAX),
  controlMaxSteps(20),
  // real vars
  comm(c),
  blockTH(true),
  expmode(false)
{
  if (skills) {
    cerr << "second AdvSkills object created, exiting" << endl;
    exit(-1);
  }
  skills = this;
}


void AdvSkills::readjustToPlayer(const WorldPlayer& my)
{
  kickRange = my.size + BALL_SIZE + 0.8 * my.kickableMargin;
  prepareOuterLimit = 3*kickRange;
  prepareInnerLimit = 2*kickRange;
  innerSafeCircle = my.size + BALL_SIZE + 0.3 * my.kickableMargin;
  safeCircle = my.size + BALL_SIZE + 0.55 * my.kickableMargin;
  dribbleMaxV = my.vmax;
}


/********************************* the AdvSkills ****************************/


/************************************ kick *********************************/
#include "AdvSkillsKick.C"

/********************************** intercept ******************************/
#include "AdvSkillsIntercept.C"

/*********************************** dribble *******************************/
#include "AdvSkillsDribble.C"

void AdvSkills::basicDash (double d) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendDash(d);
}

void AdvSkills::basicKick (double a,double b) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendKick(a,b);
}

void AdvSkills::basicTurn (double d) {
  turnParam = d;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendTurn(d);
}

void AdvSkills::basicCatch (double d) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendCatch(d);
}

void AdvSkills::noAction(void) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->noAction();
}

void AdvSkills::attentionTo(int pNum, bool ourTeam)
{
  comm->sendAttentionTo(pNum, ourTeam);
}

void AdvSkills::attentionToOff(void)
{
  comm->sendAttentionToOff();
}

void AdvSkills::basicPointTo(double dist, double dir) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendPointTo(dist, dir);
}

void AdvSkills::basicPointToOff(void) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendPointToOff();
}

void AdvSkills::basicTackle(double power) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendTackle(power);
}

void AdvSkills::basicMove(double x, double y) {
  turnParam = 0;
  turnHeadParam = 0;
  blockTH = false;
  comm->sendMove(x,y);
}

void AdvSkills::sendTurnNeck(double dir)
{
  turnHeadParam = dir;
  comm->sendTurnNeck(dir);
}

void AdvSkills::sendChangeView(view_width width, view_quality quality)
{
  comm->sendChangeView(width, quality);
}

void AdvSkills::sendSenseBody()
{
  comm->sendSenseBody();
}

/*********************************** catch *******************************/

bool AdvSkills::catchSafe(const Snapshot* sn) 
{
  
  return sn->my.ballCatchable (sn->ball, 
			       (sn->ball.pos - sn->my.pos).getAngle()
			       - sn->my.bodyDir );
}

AdvSkills::CatchCommand AdvSkills::goalieCatch(const Snapshot *sn)
{
  CatchCommand com;
  com.catchDir = (sn->ball.pos - sn->my.pos).getAngle();
  // check if player is goalie and if ball is in catchable area
  if(!comm->amIGoalie()) {
    com.state = CNOTGOALIE;
    return com;
  }
  if ((sn->time - sn->lastCatchTime) < CATCH_BAN_CYCLES) {
    com.state = CBAN;   
    return com;
  }
  if (!catchSafe(sn)) {
    com.state = CRANGE;
    return com;
  }
  com.state = COK;
  return com;
}

void AdvSkills::exec(const Snapshot *sn, const AdvSkills::CatchCommand &com) {
  basicCatch(com.catchDir - sn->my.bodyDir);
}

/*********************************** turn *******************************/
AdvSkills::TurnCommand AdvSkills::turn(const Snapshot *sn, Angle a)
{
  TurnCommand com;

  // NOT an angle since we need to know if we turn to hard (inertiaMoment)
  double delta = double(a - sn->my.bodyDir) * sn->my.inertiaMoment();
  com.state = (fabs(delta) < turnMinAngle) ? RFINISHED : ROK;

  if (delta >  M_PI * MAXMOMENT / 180.0) {
    delta = M_PI * MAXMOMENT / 180.0;
    com.state = RANGLE;
  }

  if (delta < M_PI * MINMOMENT / 180.0) {
    delta = M_PI * MINMOMENT / 180.0;
    com.state = RANGLE;
  }

  com.turnMoment = delta;
  com.bodyDir = sn->my.bodyDir +
    Angle(com.turnMoment / sn->my.inertiaMoment());
  return com;
}

void AdvSkills::exec(const Snapshot *, const AdvSkills::TurnCommand &com) { 
  basicTurn(com.turnMoment);
}

/***************************** turnHead *******************************/

AdvSkills::TurnHeadCommand AdvSkills::turnHead(const Snapshot *sn, Angle a)
{
  TurnHeadCommand com;
  com.state = THOK;
  com.vwidth = VIEW_WIDTH_NORMAL;

  if (blockTH) {
    cerr << "call of Skills::turnHead before command set" << endl;
    com.state = THBAN;
    return com;
  }

  THNAMEDLINE("W", sn->my.pos, radAng2Vector2d(1, a));

  Angle myNextBodyDir = sn->my.bodyDir + Angle(turnParam/sn->my.inertiaMoment());
#ifdef TRACE_TURNHEAD
  if (fabs(turnParam) > 0)
    THNAMEDLINE("H", sn->my.pos, radAng2Vector2d(1, myNextBodyDir));
#endif

  a -= myNextBodyDir;
  
  if (double(a) > M_PI * MAXNECKANG / 180.0) {
    a = M_PI * MAXNECKANG / 180.0;
    com.state = THANGLE;
  }
  else if (double(a) < M_PI * MINNECKANG / 180.0) {
    a = M_PI * MINNECKANG / 180.0;
    com.state = THANGLE;
  }

  // NOT an angle, since we need to know if we turn too hard
  double delta = double(a) - double(sn->senseBody.headAngle);
  
  if ((com.state == THOK) && (fabs(delta) < turnMinAngle)) {
    com.state = THFINISHED;
    THDRAWTRACE(MLog("turnHead -> finished"));
  }

  if (delta > M_PI * MAXNECKMOMENT / 180.0) {
    delta = M_PI * MAXNECKMOMENT / 180.0;
    com.state = THANGLE;
    THDRAWTRACE(MLog("turnHead -> angle 1"));
  }
  else if (delta < M_PI * MINNECKMOMENT / 180.0) {
    delta = M_PI * MINNECKMOMENT / 180.0;
    com.state = THANGLE;
    THDRAWTRACE(MLog("turnHead -> angle 2"));
  }

  THNAMEDLINE("F", sn->my.pos, radAng2Vector2d(1, a + myNextBodyDir));

  com.faceDir = sn->my.faceDir + Angle(delta);
  com.turnHeadMoment = delta;
#ifdef TRACE_TURNHEAD
  { ostrstream tmp;
  tmp << "th req=" << a << " delta= " << delta << endl;
  char * str=tmp.str();
  THDRAWTRACE(MLog(str));  
  delete str; }
#endif
  return com;
}

AdvSkills::TurnHeadCommand
AdvSkills::turnHead(const Snapshot *sn, const WorldObject &obj)
{
  // [!] remember dashes and take them into account here 
  return turnHead(sn, (obj.pos - sn->my.pos - sn->my.v).getAngle());
}

void AdvSkills::enterExpTurnHeadTurn(double _th)
{
  if (!expmode) {
    btParam  = turnParam;
    bblockTH = blockTH;
    expmode  = true;
    blockTH  = false;
  }

  turnParam = _th;
}

void AdvSkills::leaveExpTurnHeadTurn()
{
  if (!expmode)
    return;

  turnParam = btParam;
  expmode  = false;
  blockTH  = bblockTH;
}

void AdvSkills::exec(const Snapshot *sn, const AdvSkills::TurnHeadCommand &com) {
  if (com.state == THBAN)
    return;
  
  // change of view width
  if (com.vwidth != sn->senseBody.viewWidth)
    comm->sendChangeView(com.vwidth, VIEW_QUALITY_HIGH);
  
  comm->sendTurnNeck(com.turnHeadMoment);
}

/*********************************** pointto ******************************/

AdvSkills::PointToCommand
AdvSkills::pointTo(const Snapshot *sn, const Vector2d &pos)
{
  PointToCommand com;
  WorldPlayer my(sn->my);
  com.off = false;
  com.dist = norm(pos - my.pos);
  com.dir  = (pos - my.pos).getAngle() - my.faceDir - 
    Angle(turnParam) -  Angle(turnHeadParam);

  PTTRACE(MDPosition(my.pos +
		     radAng2Vector2d(com.dist,(pos - my.pos).getAngle())));

  if ((sn->time - sn->lastPointToTime) < POINTTO_BAN_CYCLES) {
    com.state = PTBAN;   
    return com;
  }
  com.state = PTOK;
  return com;
}

AdvSkills::PointToCommand
AdvSkills::pointToOff(const Snapshot *sn)
{
  PointToCommand com;
  com.off = true;
  if ((sn->time - sn->lastPointToTime) < POINTTO_BAN_CYCLES) {
    com.state = PTBAN;   
    return com;
  }
  com.state = PTOK;
  return com;
}

void AdvSkills::exec(const Snapshot *, const AdvSkills::PointToCommand &com) {
  if (com.state == PTBAN)
    return;
  
  if (com.off)
    basicPointToOff();
  else
    basicPointTo(com.dist, com.dir);
}

/*********************************** tackle ******************************/
double
AdvSkills::tackleProb(const Snapshot *sn) {
  Vector2d p2b(sn->ball.pos - sn->my.pos);
  p2b.turn(-sn->my.bodyDir);

  const double d(p2b.getX() > 0 ? TACKLE_DIST : TACKLE_BACK_DIST);
  const double w(TACKLE_WIDTH);
  const double dwp(power(d*w,TACKLE_EXP));

  const double P(1 - (power(p2b.getX()*w, TACKLE_EXP) +
		      power(fabs(p2b.getY())*d, TACKLE_EXP)) /
	  	     dwp);
    
  return max(0., min(1., P));
}

AdvSkills::TackleCommand
AdvSkills::tackle(const Snapshot *sn)
{
  TackleCommand com;

  com.state = TOK;

  if (kickSafe(sn)) {
    com.state = TKICKABLE;
    return com;
  }

  const WorldBall &ball(sn->ball);
  const WorldPlayer &my(sn->my);

  com.failProb = 1-tackleProb(sn);

  if (com.failProb > .5) {
    com.state = TRANGE;
    return com;
  }

  com.power = MAXPOWER/TACKLE_POWER_RATE;

  Vector2d dir;
  dir.setRadAng(MAXPOWER, my.bodyDir);

  com.dir = ball.v+dir;

  double nrm(norm(com.dir));
  if (nrm) 
    com.dir.stretch(1./nrm);

  return com;
}

AdvSkills::TackleCommand
AdvSkills::tackle(const Snapshot *sn, const double &dist, const double &tgt_v)
{
  TackleCommand com;

  com.state = TOK;

  if (kickSafe(sn)) {
    com.state = TKICKABLE;
    return com;
  }

  const WorldBall &ball(sn->ball);
  const WorldPlayer &my(sn->my);

  com.failProb = 1-tackleProb(sn);

  if (com.failProb > .5) {
    com.state = TRANGE;
    return com;
  }

  double tgtv = max(tgt_v, 0.);

  Vector2d dir, v(ball.v), tgt(ball.v);
  dir.setRadAng(1., my.bodyDir);

  double p, r;
  p = v*dir;
  r = sqr(p) - (sqrnorm(v) - sqr(dist));

  double x1(-p);
  if (r >= 0)
    x1 += sqrt(r);
  else {
    com.state = TTOOFAST;
    x1 += fabs(p); // assume dist == |v|
  }
  tgt += x1*dir;

  TACDRWTRACE(MDPosition(ball.pos+tgt));

  double nrm(norm(tgt));
  if (nrm) {
    com.dir = (1./nrm)*tgt;

    tgtv = tgtv + (1-BALL_DECAY)*nrm;
    if (tgtv > BALL_SPEED_MAX) {
      tgtv = BALL_SPEED_MAX;
      com.state = TTOOSHORT;
    }
    tgt.stretch(tgtv/nrm);
  } else {
    double nrmv(norm(v));
    com.dir = nrmv ? (-1./nrmv)*v : v;
  }

  TACDRWTRACE(MDCircle(ball.pos,dist));
  TACDRWTRACE(MDLine(ball.pos,ball.pos+v));
  TACDRWTRACE(MDLine(ball.pos,ball.pos+tgt-v));

  com.power = norm(tgt-v);
  if (x1 >= 0) {
    if (com.power > MAXPOWER)
      com.power = MAXPOWER;
  }
  else {
    com.power *= -1;
    if (com.power < MINPOWER)
      com.power = MINPOWER;
  }

  com.power /= TACKLE_POWER_RATE;

  return com;
}

void AdvSkills::exec(const Snapshot *sn, const AdvSkills::TackleCommand &com) {
  if (com.state == TRANGE) {
    cerr << sn->time << " AdvSkills::exec(TackleCommand): ball too far away to tackle! (failprob=" << com.failProb << ")" << endl;
    noAction();
    return;
  }
  if (com.state == TKICKABLE) {
    cerr << sn->time << " AdvSkills::exec(TackleCommand): ball in kickrange" << endl;
    noAction();
    return;
  }

  TACTRACE("TACKLING (" << com.power << ")");
  basicTackle(com.power);
}
