//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef Polygon_h
#define Polygon_h

#include "coordinates.h"
#include <stdio.h>


// This is a class which works correctly only with KONVEX polygons
class Polygon{
  Vector2d *p;          // Pointer to a list of points sorted against the clock
  int n;                // Number of points the polygon consists of

  // check if v is allready a point of the polygon
  // return 1 if v is a point of the polygon
  // return 0 else
  bool exists( const Vector2d v );

 public:
  // create a polygon out of a list of points sorted against the clock
  // v  : Array containig the points of the polygon sorted against the clock
  // len: Length of the array v
  Polygon( const Vector2d v[], const int len = 0 );

  // copy a polygon
  Polygon( const Polygon &P );

  // create an empty polygon
  Polygon();

  // clean up
  ~Polygon();
  
  // returns the number of points the polygon has
  int nop() const;

  // returns a point inside the polygon
  Vector2d getInside() const;

  // adds a point at the end of p
  // if c is not equal to the last
  // or the first point
  // returns 1 if added
  // else 0
  int addPoint( const Vector2d c );

  // deletes the point number no
  // 0 <= no < n
  // returns 1 if deleted
  // else 0
  int delPoint( const int no );

  // should insert a point into the polygon at the right place
  int insertPoint( const Vector2d c );

  // delete all points of the polygon and set n = 0, p = NULL;
  void clear();

  // returns a pointer to an array of points of the polygon
  // sorted against the clock
  Vector2d *points() const;

  Polygon operator=( const Polygon &P );

  // returns the intersection of two KONVEX polygons
  Polygon operator^( const Polygon &P ) const;

  // returns the intersection of two KONVEX polygons
  // faster than operator^
  Polygon operator^=( const Polygon &P );

}; //end class Polygon

#endif


