//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////


//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: asyncinfo.C,v 1.24 2002/05/09 14:10:20 cs Exp $
// 
//  asyncinfo.C - storing of asynchronously updated agent information
// 
//////////////////////////////////////////////////////////////////////


#include <signal.h>
#include <sys/time.h>
#include "asyncinfo.h"
#include "Config.h"

#undef CERR
// error trace, abbreviate error messages by commenting out #define
#define CERR(output) cerr << output << endl;

#ifdef DO_SEETEST
#define SEETEST(cmd) cmd
#warning *** AsnycCommInfo: Will do see message timing test
#else
#define SEETEST(cmd)
#endif

#define SHORT_CYCLE_WARNING 0
//////////////////////////////////////////////////////////////////////


AsyncCommInfo * AsyncCommInfo::info = NULL;

AsyncCommInfo::AsyncCommInfo( const char * _team, client_type ctype ) 
  : team(NULL),
    number(-1),
    client(ctype),
    neckTwisted(false),
    serverTime(0),
    playMode(PLAY_MODE_UNKNOWN),
    side(SIDE_UNKNOWN),
    ourScore(0),
    otherScore(0),
    secondHalf(false),
    kickCount(0),
    dashCount(0),
    turnCount(0),
    sayCount(0),
    turnNeckCount(0),
    moveCount(0),
    catchCount(0),
    changeViewCount(0),
    pointToCount(0),
    attentionToCount(0),
    tackleCount(0),
    seeMsgCntIndex(-1),
    viewModeTime(0),
    viewModeUnknown(false),
    viewModeChangedAgain(false),
    curVisAngle(VIEW_WIDTH_NARROW),
    curVisQual(VIEW_QUALITY_HIGH),
    chVisAngle(VIEW_WIDTH_NARROW),
    chVisQual(VIEW_QUALITY_HIGH),
    noSenseBodyYet(true),
    actionNeeded(true),
    nextTimestep(false),
    waitForSee(false),
    act(false),
    recvTime(0),
    waitForSeeTime(40),
    abortWaitForSee(false)
{
  ASYNCTRACE(cout << "AsyncInfo: entering constructor" << endl);

  if (!info)
    info = this;
  else
    CERR("You can create only one AsyncCommInfo!");

  gettimeofday(&startTime, NULL);
  cycleStartTime.tv_sec = cycleStartTime.tv_usec = 0;

  waitForSeeTime = Config::getValueAsInt("async_waitForSeeTime");
  ensure(waitForSeeTime%10 == 0 && 0 <= waitForSeeTime && waitForSeeTime <= 90);

  team = new char[strlen(_team)+1];
  strcpy(team, _team);

  clearAction(serverTime, sentAction);
  processedAction = lostAction = sentAction;

  // setup signalhandling
  sigset_t sset;
  sigemptyset(&sset);
  sigaddset(&sset, SIGALRM);

  struct sigaction saction;
  saction.sa_handler = alrmHandler;
  saction.sa_flags   = SA_RESTART;
  saction.sa_mask    = sset;
  sigaction(SIGALRM, &saction, NULL);

#ifdef DO_SEETEST
  sCnt       = 0;
  sGotCnt    = 0;
  diff       = 0;
  sStartTime = 0;
  sDiff      = 0.;
#endif

  seeMsgCnt[0] = seeMsgCnt[1] = seeMsgCnt[2] = seeMsgCnt[3] = 0;
  seeTime[0] = seeTime[1] = seeTime[2] = seeTime[3] = 0;

  ASYNCTRACE(cout << "AsyncInfo: leaving constructor" << endl);
}


AsyncCommInfo::~AsyncCommInfo() {
  ASYNCTRACE(cout << "AsyncInfo: entering destructor" << endl);
  SEETEST(if (!isCoach()) cerr << "(" << team << " No. " << number << ") average time of seemsgs: "
	  << sDiff/sCnt << " msec. Waiting for " << waitForSeeTime << " msec." << endl 
	  << "            Got " << sGotCnt << " out of "
	  << sCnt << " msgs " << 100*double(sGotCnt)/sCnt << "%" << endl);
  delete[] team;
  ASYNCTRACE(cout << "AsyncInfo: leaving destructor" << endl);
}


void
AsyncCommInfo::setStates( int _number, int _time, play_mode _mode, 
			  playing_side _side, bool _secondHalf )
{
  number     = _number;
  serverTime = _time;
  playMode   = _mode;
  side       = _side;
  secondHalf = _secondHalf;

#ifndef CERR
#define CERR(output) cerr << "A";
#warning AsyncInfo error trace is abbreviated to A
  cerr << "*** Warning: AsyncInfo(" 
       << team << " No." << number 
       << "): error trace is abbreviated to A, cf. source" << endl;
#endif
}

void 
AsyncCommInfo::setAction( CycleCommand cmd, int time, 
			  double param1, double param2,
			  CycleAction &action )
{
  action.command = cmd;
  action.time = time;
  action.param1 = param1;
  action.param2 = param2;
}

void
AsyncCommInfo::clearAction( int time, CycleAction &action )
{
  setAction(NO_COMMAND, time, 0, 0, action);
  ASYNCTRACE(cout << "           marking action as unprepared" << endl);
}

void
AsyncCommInfo::getLastAction( CycleAction &action )
{
  ASYNCTRACE(cout << serverTime << " AsyncInfo: returning last action (of time " 
	<< processedAction.time << ") = ";
	switch (processedAction.command) {
	case NO_COMMAND   : cout << "none" 
				 << endl; break;
	case KICK_COMMAND : cout << "kick " 
				 << processedAction.param1 << " " 
				 << processedAction.param2 << endl; break;
	case DASH_COMMAND : cout << "dash " 
				 << processedAction.param1 << endl; break;
	case TURN_COMMAND : cout << "turn " 
				 << processedAction.param1 << endl; break;
	case CATCH_COMMAND: cout << "catch " 
				 << processedAction.param1 << endl; break;
	case MOVE_COMMAND : cout << "move " 
				 << processedAction.param1 << " "
				 << processedAction.param2 << endl; break;
	case TACKLE_COMMAND : cout << "tackle " 
				   << processedAction.param1
				   << endl; break;
	default:;
	});

  action = processedAction;
}

void 
AsyncCommInfo::newTimestep( void )
{
  // set flag, so that checking of command counters will take place after
  // next sense_body, and new command will be requested from agent.
  noSenseBodyYet = true;

  // assume sending went ok until checking of counters can be performed
  if (lostAction.command == NO_COMMAND)
    processedAction = sentAction;
  else
    processedAction = lostAction;
 
  abortWaitForSee = false;
  if (!isCoach()) // coach does not have to wait for see info
    startTimer();

  nextTimestep = true;

  if (seeMsgCntIndex <= 2) 
    ++seeMsgCntIndex;
  else
    seeMsgCntIndex = 0;

  seeMsgCnt[seeMsgCntIndex] = seeTime[seeMsgCntIndex] = 0;

  ++viewModeTime;

  updateAct();
}

void
AsyncCommInfo::setRecvTime( void )
{
  struct timeval now;
  gettimeofday(&now, NULL);
  now.tv_sec -= startTime.tv_sec;
  recvTime = tv2msec(now);
  SEETRACE(cout << serverTime << " AsyncInfo: recvd msg at: " << recvTime << endl);
}

void
AsyncCommInfo::startTimer( void )
{
  struct timeval now;
  struct itimerval tv;
  int maxTime = int(2*SIMULATOR_STEP-waitForSeeTime/100.*SIMULATOR_STEP+10);

  // we want to have a single shot timer
  tv.it_interval.tv_sec = 0;
  tv.it_interval.tv_usec = 0;

  gettimeofday(&now,NULL);
  now.tv_sec -= startTime.tv_sec;

  long t(tv2msec(now));

  SEETRACE(cout << serverTime << " AsyncInfo: start timer at: " << t << endl);
  SEETEST(diff = -1; sStartTime = t);

  long wait = long(waitForSeeTime/100.*SIMULATOR_STEP-(t-recvTime));
  long cycleTime(t-tv2msec(cycleStartTime));
  msec2tv(recvTime, cycleStartTime);

  SEETRACE(cout << serverTime << " AsyncInfo: cycle starts at: " 
	   << tv2msec(cycleStartTime) << endl);

  if (cycleTime < SHORT_CYCLE_WARNING*SIMULATOR_STEP) {
      CERR(serverTime << " AsyncInfo: (" 
	   << getTeam() << " No." << getNumber() << ")"
	   << " last cycle was to short: " << cycleTime << " ("
	   << SIMULATOR_STEP << ")");
  }

  if (wait > 0 && cycleTime <= maxTime) {

    if (!viewModeUnknown && !viewModeChangedAgain &&
	curVisQual == chVisQual && curVisAngle == chVisAngle) {
      if (viewModeTime >= 2) {
	if (curVisAngle == VIEW_WIDTH_NORMAL && 
	    curVisQual == VIEW_QUALITY_HIGH ||
	    curVisAngle == VIEW_WIDTH_WIDE && 
	    curVisQual == VIEW_QUALITY_LOW) {
	  if (seeMsgCnt[seeMsgCntIndex] == 1 &&
	      seeMsgCnt[(seeMsgCntIndex+3)%4] == 1) {
	    wait = 0;
	    abortWaitForSee = true;
	    SEETRACE(cout << "not starting timer as I wont get a see msg" << endl);
	  }
	  if (seeMsgCnt[seeMsgCntIndex] == 0 &&
	      seeMsgCnt[(seeMsgCntIndex+3)%4] == 1) {
	    wait = min(wait, (seeTime[(seeMsgCntIndex+3)%4]+
			      2*SEND_VI_STEP+5)%101);
	  }

	  if (seeMsgCnt[seeMsgCntIndex] == 1 &&
	      seeMsgCnt[(seeMsgCntIndex+3)%4] == 0) {
	    wait = min(wait, (seeTime[seeMsgCntIndex]+
			      2*SEND_VI_STEP+5)%101);
	  }
	}

	if (curVisAngle == VIEW_WIDTH_NARROW && 
	    curVisQual == VIEW_QUALITY_HIGH) {
	  if (seeMsgCnt[seeMsgCntIndex] == 1 &&
	      seeMsgCnt[(seeMsgCntIndex+3)%4] == 1)
	    wait = min(wait, (seeTime[seeMsgCntIndex]+
			      2*SEND_VI_STEP+5)%101);

	  if (seeMsgCnt[seeMsgCntIndex] == 2 ||
	      seeMsgCnt[seeMsgCntIndex] == 1 &&
	      seeMsgCnt[(seeMsgCntIndex+3)%4] == 2)
	    wait = min(wait, (seeTime[seeMsgCntIndex]+
			      SEND_VI_STEP+5)%101);
	}

	if (curVisAngle == VIEW_WIDTH_WIDE && 
	    curVisQual == VIEW_QUALITY_HIGH) {
	  if (seeMsgCnt[seeMsgCntIndex] == 1 ||
	      seeMsgCnt[(seeMsgCntIndex+3)%4] == 1) {
	    wait = 0;
	    abortWaitForSee = true;
	    SEETRACE(cout << "not starting timer as I wont get a see msg" << endl);
	  }
	}
      }
    }

    // we want to wait for audiomsgs
    wait = max(10,wait);
    msec2tv(wait, tv.it_value);
    setitimer(ITIMER_REAL, &tv, NULL);
    SEETRACE(cout << "started timer with " << tv.it_value.tv_sec 
	     << " seconds and " << tv.it_value.tv_usec << " useconds" << endl);
  }
  else {
    abortWaitForSee = true;
    msec2tv(0, tv.it_value); // stop eventually running timer
    setitimer(ITIMER_REAL, &tv, NULL);
    if (waitForSeeTime > 0 && wait <= 0) {
      CERR(serverTime << " AsyncInfo: (" 
	   << getTeam() << " No." << getNumber() << ")"
	   << " aborting waitForSee due to non positive timervalue ("
	   << wait << ")");
    }
    if (cycleTime > maxTime && serverTime > 0 && waitForSeeTime > 0) {
      CERR(serverTime << " AsyncInfo: ("
	   << getTeam() << " No." << getNumber() << ")"
	   << " timer not started as last cycle was to long: "
	   << cycleTime << " (" << maxTime << ")");
    }
  }
  updateAct();
}

void 
AsyncCommInfo::checkSyncCommandCount( const SensorData &data )
{
  ASYNCTRACE (cout << serverTime << " AsyncInfo: checking for lost commands:" << endl
	 << "           old (kick, dash, turn, move, catch) = " 
	 << kickCount  << " " 
	 << dashCount  << " " 
	 << turnCount  << " " 
	 << moveCount  << " " 
	 << catchCount  << " " 
	 << tackleCount  << " " 
	 << "           current action = ";
	 switch (sentAction.command) {
	 case NO_COMMAND   : cout << "none"; break;
	 case KICK_COMMAND : cout << "kick"; break;
	 case DASH_COMMAND : cout << "dash"; break;
	 case TURN_COMMAND : cout << "turn"; break;
	 case MOVE_COMMAND : cout << "move"; break;
	 case CATCH_COMMAND : cout << "catch"; break;
	 case TACKLE_COMMAND : cout << "tackle"; break;
	 default:;
	 }
	 cout << endl << "           rcv (kick, dash, turn, move, catch) = "
	 << data.kick_count  << " " 
	 << data.dash_count  << " " 
	 << data.turn_count  << " " 
	 << data.move_count  << " " 
	 << data.catch_count  << " " << endl);
  
  // check for missed kick, dash, turn, ...

  CycleCommand execCmd = NO_COMMAND;

  // find the command which was executed by the server
  if (data.kick_count - kickCount > 0) 
    execCmd = KICK_COMMAND;
  if (data.dash_count - dashCount > 0) {
    if (execCmd != NO_COMMAND) 
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): Server executed more than one command");
    execCmd = DASH_COMMAND;
  } 
  if (data.turn_count - turnCount > 0) {
    if (execCmd != NO_COMMAND) 
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): Server executed more than one command");
    execCmd = TURN_COMMAND;
  }
  if (data.move_count - moveCount > 0) {
    if (execCmd != NO_COMMAND) 
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): Server executed more than one command");
    execCmd = MOVE_COMMAND;
  }
  if (data.catch_count - catchCount > 0) {
    if (execCmd != NO_COMMAND) 
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): Server executed more than one command");
    execCmd = CATCH_COMMAND;
  }
  if (data.tackle_count - tackleCount > 0) {
    if (execCmd != NO_COMMAND) 
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): Server executed more than one command");
    execCmd = TACKLE_COMMAND;
  }


  if (execCmd == NO_COMMAND) {
    /* No command was executed by the server, so set the processed
       action to none and lost action to the command, which was sent.
       Then a new one is calculated by the agent and sent to the server.
     */

    processedAction.command = NO_COMMAND;

    switch (sentAction.command) {
    case KICK_COMMAND:
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): missing " << kickCount+1 - data.kick_count
	    << " kick command(s)!" );
      lostAction = sentAction;
      break;
    case DASH_COMMAND:
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): missing " << dashCount+1 - data.dash_count
	    << " dash command(s)!" );
      lostAction = sentAction;
      break;
    case TURN_COMMAND:
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): missing " << turnCount+1 - data.turn_count
	    << " turn command(s)!" );
      lostAction = sentAction;
      break;
    case MOVE_COMMAND:
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): missing " << moveCount+1 - data.move_count
	    << " move command(s)!" );
      lostAction = sentAction;
      break;
    case CATCH_COMMAND:
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): missing " << catchCount+1 - data.catch_count
	    << " catch command(s)!" );
      lostAction = sentAction;
    case TACKLE_COMMAND:
      CERR( serverTime 
	    << " AsyncInfo (" 
	    << getTeam() << " No." << getNumber()
	    << "): missing " << tackleCount+1 - data.tackle_count
	    << " tackle command(s)!" );
      lostAction = sentAction;
      break;
    default:;
    }
    ASYNCTRACE( cout << "Parser: lost command is: ";
		switch (lostAction.command) {
		case NO_COMMAND   : cout << "none"; break;
		case KICK_COMMAND : cout << "kick"; break;
		case DASH_COMMAND : cout << "dash"; break;
		case TURN_COMMAND : cout << "turn"; break;
		case MOVE_COMMAND : cout << "move"; break;
		case CATCH_COMMAND : cout << "catch"; break;
		case TACKLE_COMMAND : cout << "tackle"; break;
		default:;
		}
		cout << endl;);

  } else { // server executed a command
    /* Assuming that commands are only delayed and not lost
       i.e. as only the first cmd per cycle is executed by 
       the server we should not resend the command but calc
       a new one which is sent to the server. Then check wich
       one was executed */
    if (execCmd == lostAction.command)
      processedAction = lostAction;
    else if (execCmd == sentAction.command)
      processedAction = sentAction;
    else {
      processedAction.command = NO_COMMAND;
      cerr << serverTime 
	   << " AsyncInfo (" 
	   << getTeam() << " No." << getNumber()
	   << "): Server executed command ("; 
      switch (execCmd) {
      case KICK_COMMAND : cerr << "kick"; break;
      case DASH_COMMAND : cerr << "dash"; break;
      case TURN_COMMAND : cerr << "turn"; break;
      case MOVE_COMMAND : cerr << "move"; break;
      case CATCH_COMMAND : cerr << "catch"; break;
      case TACKLE_COMMAND : cerr << "tackle"; break;
      default: cout << execCmd << endl; break;
      }
      cerr << ") I didn't send!" << endl;
    }

    clearAction(serverTime, lostAction);
  }
  
  ASYNCTRACE(cout << "Parser: processedAction: ";
	     switch (processedAction.command) {
	     case NO_COMMAND   : cout << "none"; break;
	     case KICK_COMMAND : cout << "kick"; break;
	     case DASH_COMMAND : cout << "dash"; break;
	     case TURN_COMMAND : cout << "turn"; break;
	     case MOVE_COMMAND : cout << "move"; break;
	     case CATCH_COMMAND : cout << "catch"; break;
	     case TACKLE_COMMAND : cout << "tackle"; break;
	     default:;
	     }
	     cout << endl;);

  clearAction(serverTime, sentAction);
  actionNeeded = true;
  updateAct();

  // after printing error messages: update counters
  kickCount = data.kick_count;
  dashCount = data.dash_count;
  turnCount = data.turn_count;
  moveCount = data.move_count;
  catchCount = data.catch_count;
  tackleCount = data.tackle_count;
}

void 
AsyncCommInfo::checkAsyncCommandCount(const SensorData &data) {

  ASYNCTRACE (cout << serverTime << " AsyncInfo: checking for lost commands:" << endl
	 << "           old (say, turn_neck, changeView, pointTo) = " 
	 << sayCount << " " 
	 << turnNeckCount << " "
	 << changeViewCount << " " 
	 << pointToCount
	 << endl;
	 cout << "           rcv (say, turn_neck, changeView) = "
	 << data.say_count << " " 
	 << data.turn_neck_count << " " 
	 << data.change_view_count << " "
	 << data.point_to_count
	 << endl);
  
  // check for missed say, turn_neck, etc.
  if (sayCount != data.say_count) {
    CERR( serverTime 
	 << " AsyncInfo (" << getTeam() << " No." << getNumber()
	 << "): missing say commands "
	 << "(sent " << sayCount 
	 << ", was told " << data.say_count << ")"  );
    sayCount = data.say_count;
  }
  if (turnNeckCount != data.turn_neck_count) {
    CERR( serverTime 
	 << " AsyncInfo (" << getTeam() << " No." << getNumber()
	 << "): missing turn_neck commands "
	 << "(sent " << turnNeckCount 
	 << ", was told " << data.turn_neck_count << ")"  );
    turnNeckCount = data.turn_neck_count;
  }
  if (changeViewCount != data.change_view_count) {
    CERR( serverTime 
	 << " AsyncInfo (" << getTeam() << " No." << getNumber()
	 << "): missing change view commands "
	 << "(sent " << changeViewCount 
	 << ", was told " << data.change_view_count << ")"  );

    if (changeViewCount < data.change_view_count) {
      if (viewModeChangedAgain) {
	viewModeUnknown = true;
      } else {
	viewModeUnknown = false;
	curVisAngle = chVisAngle;
	curVisQual = chVisQual;
      }
      viewModeTime = 0;
    } 

    changeViewCount = data.change_view_count;
  }
  else {
    if ((curVisAngle != chVisAngle || curVisQual != curVisQual)
	&& !viewModeChangedAgain)
      {
	curVisQual = chVisQual;
	curVisAngle = chVisAngle;
	viewModeTime = 0;
      }
  }

  if (pointToCount != data.point_to_count) {
    CERR( serverTime 
	 << " AsyncInfo (" << getTeam() << " No." << getNumber()
	 << "): missing point to commands "
	 << "(sent " << pointToCount 
	 << ", was told " << data.point_to_count << ")"  );
    pointToCount = data.point_to_count;
  }
  
  if (attentionToCount != data.attention_to_count) {
    CERR( serverTime 
	 << " AsyncInfo (" << getTeam() << " No." << getNumber()
	 << "): missing attention to commands "
	 << "(sent " << attentionToCount 
	 << ", was told " << data.attention_to_count << ")"  );
    attentionToCount = data.attention_to_count;
  }
  
}

bool
AsyncCommInfo::adjustTime( const SensorData &data )
{
  ASYNCTRACE(cout << "AsyncInfo: current time = " << serverTime 
	 << ", adjusting to data.time = " << data.time << endl);

  if (data.time > serverTime) {  // new data

    if (data.time > serverTime + 1) { // data MUCH TOO new

      CERR( data.time
	   << " AsyncInfo (" << getTeam() << " No." << getNumber()
	   << "): completely missed server timesteps from "
	   << serverTime+1 << " to " << data.time-1 << " (no commands sent)" 
	    );

    }
    else if ( actionNeeded && 
	      !isCoach() &&
	      isTimeRunning() ) {
      
      // isTimeRunning() is appropriate here for the lost command check:
      
      // if game was stopped, the play mode will still reflect this right now, 
      // as it is updated AFTER adjustTime() in adjust(). So: prompt error
      // only if game was already running -- to avoid error prompts when
      // game is about to be restarted.

      CERR( serverTime
	   << " AsyncInfo (" << getTeam() << " No." << getNumber()
	   << "): agent did not send command"  );
      ensure(false);
    }

    SENSETRACE(switch(data.type) {
	       case SENSOR_TYPE_SEE:
		 cout << serverTime << "AsyncInfo: see was first message !" << endl; break;
	       
               case SENSOR_TYPE_HEAR:
		 cout << serverTime << "AsyncInfo: hear was first message !" << endl; break;
	       
               case SENSOR_TYPE_SENSE_BODY:
		 break; // was ok
	       
               default:
		 cout << serverTime << "AsyncInfo: " << (int)data.type 
		      << "was first message !" << endl; break;
	       });

    serverTime = data.time;
    newTimestep();
    return true;
  } 
  else {
    return false;
  }
}

bool 
AsyncCommInfo::adjust(const SensorData &data) {

  bool firstSenseBody(false);

  // 1. adjust server time according to data
  adjustTime(data);
  

  // 2. check for new timestep for coach when clock is off
  if (data.type == SENSOR_TYPE_EXACT && 
      !isTimeRunning()) {
    newTimestep();
  }

  // 3a. adjust headAngle, 
  // 3b. force new timestep for player when clock is off,
  // 3c. check and update command counters
  if (data.type == SENSOR_TYPE_SENSE_BODY) {

    // we KNOW our own head_angle after receiving a sense_body
    neckTwisted = false;

    
    if (!isTimeRunning() && 
	!noSenseBodyYet) { // force new timestep when clock is stopped.
      ASYNCTRACE(cout << serverTime << " AsyncInfo: game is stopped, "
	     << "had a sense_body -> \"next\" timestep !" << endl);
      newTimestep();
      // This might confuse the action counters temporarily (for one step)
      // when additional sense_body
      // messages are received while clock is stopped, e.g. through turn_necks
      // etc. Just avoid receiving additional (sense_body *) in this case.
    }

    
    if (noSenseBodyYet) { 
      // upon first sense_body received in a timestep:
      
      // clear flag
      noSenseBodyYet = false;
      // check turn, dash, kick (only once per step)
      checkSyncCommandCount(data); 
      // set return value
      firstSenseBody = true;
      updateAct();
    } 
  
    // 4. check say, turn_neck for every sense_body
    checkAsyncCommandCount(data); 

  }

  if (data.type == SENSOR_TYPE_SEE) {
    struct timeval now;
    gettimeofday(&now, NULL);
    now.tv_sec -= startTime.tv_sec;

    ++seeMsgCnt[seeMsgCntIndex];
    seeTime[seeMsgCntIndex] = tv2msec(now) - tv2msec(cycleStartTime);
    

    SEETEST(if (!isCoach() && diff == -1 && 
		(diff = (tv2msec(now)-sStartTime)) >=  0) {
               sDiff += diff;
               ++sCnt;
               if (!abortWaitForSee) ++sGotCnt;
            });

  }

  SEETRACE(cout << "see msgs in last 4 cycles: "
	   << seeMsgCnt[(seeMsgCntIndex+1)%4] << " " 
	   << seeMsgCnt[(seeMsgCntIndex+2)%4] << " "
	   << seeMsgCnt[(seeMsgCntIndex+3)%4] << " "
	   << seeMsgCnt[seeMsgCntIndex] << endl
	   << "in those cycles last see msg arrived at: "
	   << seeTime[(seeMsgCntIndex+1)%4] << " " 
	   << seeTime[(seeMsgCntIndex+2)%4] << " "
	   << seeTime[(seeMsgCntIndex+3)%4] << " "
	   << seeTime[seeMsgCntIndex] << endl
	   << "viewModeTime: " << viewModeTime << endl
	   << "viewModeUnknown: " << viewModeUnknown << endl
	   << "viewModeChangedAgain: " << viewModeChangedAgain << endl
	   );

  // adjust playMode according to referee message
  if ( (data.type == SENSOR_TYPE_HEAR) ) {
    if (data.sender == SENDER_REFEREE) {
      /* Nodaism! When the time is stoped a new timestep starts 
	 when the playmode changes execept we are before kick off 
	 as we do not get a sense_body then.
       */
      if (!isTimeRunning() && serverTime > 0) {
	ASYNCTRACE(cout << serverTime << " AsyncInfo: time not running and "
		   << "playmode changes --> newTimestep"
		   << endl);
	newTimestep();
      }
      playMode = data.mode;
      ASYNCTRACE(cout << serverTime << " AsyncInfo: setting Playmode to "
		 << playMode << endl);
      if (playMode == PLAY_MODE_GOAL_US) ourScore = data.score;
      if (playMode == PLAY_MODE_GOAL_THEM) otherScore = data.score;
      if (playMode == PLAY_MODE_HALF_TIME) secondHalf = true;
    }
  }

  return firstSenseBody;
}
