//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: asyncinfo.h,v 1.16 2002/05/28 15:26:37 cs Exp $
// 
//  asyncinfo.h - storing of player related asynchronous information
// 
//////////////////////////////////////////////////////////////////////


#ifndef _ASYNCINFO_H_
#define _ASYNCINFO_H_

#include <string.h>
#include <iostream.h>
#include <sys/time.h>
#include "sensor.h"

#undef ASYNCTRACE
#undef SENSETRACE
#undef SEETRACE
#undef SEETEST
#undef DO_SEETEST

#define ASYNCTRACE(cmd)
#define SENSETRACE(cmd)
#define SEETRACE(cmd)
//#define DO_SEETEST

/////////////////////////////////////////////////////////////////////////////

class AsyncCommInfo {
  
 public:

  AsyncCommInfo( const char* _team, client_type ctype );
  
  ~AsyncCommInfo( void );
  
  inline const char *  getTeam( void ) { return team; };
  inline int getNumber( void ) { return number; };
  inline bool isOnlineCoach( void ) { return client == CLIENT_ONLINECOACH; };
  inline bool isTrainCoach( void ) { return client == CLIENT_TRAINCOACH; };
  inline bool isCoach( void ) { return isOnlineCoach() || isTrainCoach(); };
  inline bool isGoalie( void ) { return client == CLIENT_GOALIE; };
  inline bool isNeckTwisted( void ) { return neckTwisted; };

  void setRecvTime( void );
  inline int getServerTime( void ) { return serverTime; };
  inline play_mode getPlayMode( void ) { return playMode; };
  inline playing_side getSide( void ) { return side; };
  inline int getOurScore( void ) { return ourScore; };
  inline int getOtherScore( void ) { return otherScore; };
  inline bool inSecondHalf( void ) { return secondHalf; };

  inline
  void twistNeck( void ) {
    ASYNCTRACE(cout << serverTime << "AsyncInfo: twisting neck" <<endl);
    neckTwisted = true;
  };

  inline
  void incSayCount( void ) {
    ASYNCTRACE(cout << serverTime << "AsyncInfo: increasing sayCount to " << sayCount <<endl);
    ++sayCount;
  };
  
  inline
  void incTurnNeckCount( void ) {
    ASYNCTRACE(cout << serverTime << "AsyncInfo: increasing turnNeckCount to " << turnNeckCount <<endl);
    ++turnNeckCount;
  };

  inline
  void incChangeViewCount( void ) {
    ASYNCTRACE(cout << serverTime << "AsyncInfo: increasing changeViewCount to " 
	   << changeViewCount <<endl);
    ++changeViewCount;
  };

  inline
  void setViewMode( view_width width, view_quality qual) {
    if (curVisAngle == chVisAngle 
	&& curVisQual == chVisQual
	|| viewModeUnknown)
      {
	chVisAngle = width;
	chVisQual  = qual;
	viewModeChangedAgain = false;
      } 
    else
      viewModeChangedAgain = true;
  };

  inline
  void incPointToCount( void ) {
    ASYNCTRACE(cout << serverTime << "AsyncInfo: increasing pointToCount to " 
	   << pointToCount << endl);
    ++pointToCount;
  };

  inline
  void incAttentionToCount( void ) {
    ASYNCTRACE(cout << serverTime << "AsyncInfo: increasing attentionToCount to " 
	   << attentionToCount << endl);
    ++attentionToCount;
  };

  inline
  bool isTimeRunning( void ) {
    return ! ( getPlayMode() == PLAY_MODE_BEFORE_KICK_OFF ||
	       getPlayMode() == PLAY_MODE_TIME_OVER ||
	       getPlayMode() == PLAY_MODE_OFFSIDE_US ||
	       getPlayMode() == PLAY_MODE_OFFSIDE_THEM ||
	       getPlayMode() == PLAY_MODE_GOAL_US ||
	       getPlayMode() == PLAY_MODE_GOAL_THEM ||
	       getPlayMode() == PLAY_MODE_BACKPASS_US ||
	       getPlayMode() == PLAY_MODE_BACKPASS_THEM ||
	       getPlayMode() == PLAY_MODE_FREEKICK_FAULT_US ||
	       getPlayMode() == PLAY_MODE_FREEKICK_FAULT_THEM ||
	       getPlayMode() == PLAY_MODE_PENALTY_KICK_US ||
	       getPlayMode() == PLAY_MODE_PENALTY_KICK_THEM ||
	       getPlayMode() == PLAY_MODE_PAUSE
	       );
  };

  inline
  void setSentAction( CycleCommand cmd, int time, 
		      double param1, double param2 )
  { 
    setAction(cmd, time, param1, param2, sentAction);
    actionNeeded = false;
    updateAct();
  };

  // same as actNow, but does not change any variables
  inline const bool * isActionNeeded( void ) { return &act; };    
  inline bool actNow( void ) {
    if (act) { 
      nextTimestep = false; 
      updateAct();
      return true;
    } 
    else 
      return false; 
  };
  void getLastAction( CycleAction &action );

  // use Sensor Data to adjust time, playMode and related states,
  // twisted neck, and action counters; and check action count.
  // returns true, if data was the first sense_body data of a timestep
  bool adjust( const SensorData &data );

  // for access to states when setting up the agent only !!
  void setStates( int _number, int _time, play_mode _mode, 
		  playing_side _side, bool _secondHalf );


  /** return the number of see msgs in the last four cycles
      @param c3 the number of see msgs in timestep minus 3
      @param c2 the number of see msgs in timestep minus 2
      @param c1 the number of see msgs in timestep minus 1
      @param c2 the number of see msgs in current cycle
   */
  inline
  void getSeeCnt( int & c3, int & c2, int & c1, int & c0 ) {
    c2 = seeMsgCnt[(seeMsgCntIndex+1)%4];
    c2 = seeMsgCnt[(seeMsgCntIndex+2)%4];
    c1 = seeMsgCnt[(seeMsgCntIndex+3)%4];
    c0 = seeMsgCnt[seeMsgCntIndex];
  };

private:

  static AsyncCommInfo * info;

  char * team;
  int number;
  client_type  client;
  bool neckTwisted;

  int serverTime;
  play_mode playMode;
  playing_side side;
  int ourScore;
  int otherScore;
  bool secondHalf;

  int kickCount;
  int dashCount;
  int turnCount;
  int sayCount;
  int turnNeckCount;
  int moveCount;
  int catchCount;
  int changeViewCount;
  int pointToCount;
  int attentionToCount;
  int tackleCount;

  // @name variables used to keep track of the view mode
  //@{  
  int seeMsgCnt[4];    // No. of see msgs in the last four cycles
  int seeTime[4];      // time in a cycle when the last see msg arrived
  int seeMsgCntIndex;  // Index to the array above
  int viewModeTime;    // how manny cycles do we have the current view mode
  bool viewModeUnknown;
  bool viewModeChangedAgain;
  view_width    curVisAngle;
  view_quality curVisQual;
  view_width    chVisAngle;
  view_quality chVisQual;
  //@}

  bool noSenseBodyYet;        // true: no sense body received in current step
  bool actionNeeded;          // true: no action sent yet in current step
  bool nextTimestep;          // true: transition was made to new step
  bool waitForSee;            // true: we have to wait for see info this step
  bool act;                   // true: we have to act in the current cycle

#ifdef DO_SEETEST
  unsigned long sCnt, sGotCnt, sStartTime;
  int diff;
  double sDiff;
#endif  

  struct timeval startTime, cycleStartTime;
  long recvTime;
  int waitForSeeTime;  // msec to wait in a new timestep for a see
  bool abortWaitForSee;
  static void alrmHandler( int ) 
  { 
    SEETRACE(cout << "alrmHandler" << endl);
    AsyncCommInfo::info->abortWaitForSee = true;
    AsyncCommInfo::info->updateAct();
  };
  void startTimer( void );

  inline
  long tv2msec( const struct timeval &tv )
  {
    return tv.tv_sec*1000+tv.tv_usec/1000;
  };

  inline
  void msec2tv( const long &msec, struct timeval &tv )
  {
    tv.tv_sec = msec/1000;
    tv.tv_usec = (msec%1000)*1000;
  };
  
  CycleAction sentAction;     // action that was just sent, unconfirmed
  CycleAction processedAction;// action processed by server from the last step
  CycleAction lostAction;     // action lost in the last step
  
  void clearAction( int time, CycleAction &action );
  void setAction( CycleCommand cmd, int time, 
		  double param1, double param2,
		  CycleAction &action );
  
  void newTimestep( void );
  bool adjustTime( const SensorData &data );

  void checkSyncCommandCount( const SensorData &data );   // turn, dash, kick
  void checkAsyncCommandCount( const SensorData &data );  // say, turn_neck

  inline void updateAct( void ) {
    act = isCoach() ? nextTimestep : 
      (actionNeeded && !noSenseBodyYet && nextTimestep && abortWaitForSee);
    SEETRACE(cout << "AsyncInfo: updateAct (" << actionNeeded << "&" 
                                              << !noSenseBodyYet << "&"
                                              << nextTimestep << "&"
	                                      << abortWaitForSee << ")" 
                                              << isCoach() << endl);
  };
};


#endif
