//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*-////////////////////////////////////////////////////////
//
// $Id: onlinecoach.C,v 1.19 2002/01/27 14:23:03 cs Exp $
//
// onlinecoach.C - interface for online coach client
//
///////////////////////////////////////////////////////////////////////

#include "onlinecoach.h"
#include "Config.h"

#undef TRACE
#define TRACE(cmd)


Onlinecoach::Onlinecoach (const char*  team_name,             
			  const char*  host,
			  int          port,
			  const char*  configfile,
			  bool         logging_snap,
                          bool         FUNSSELmode,
			  int          numSnapshots) :

  comm(NULL),
  world(NULL)
{
  TRACE (cout << "Onlinecoach: initializing config reader" << endl);

  // try to initialize config file
  if (!Config::readFile(configfile)) {
    cerr << "PANIC: Could not read config file: " << configfile << endl;
    exit(-1);
  }

  TRACE (cout << "Onlinecoach: generating Comm" << endl);
    
  comm = new Comm(team_name,
		  CLIENT_ONLINECOACH,
		  host, port,
		  0,
		  FUNSSELmode);

  comm_error ce;

  if (comm->isServerAlive()) {
    TRACE (cout << "Onlinecoach: connecting" << endl);
    ce = comm->connect();
  }
  else
    ce = COMM_ERROR;

  switch (ce) {
  case (COMM_OK) : 
      
    TRACE (cout << "Onlinecoach: generating Worldmodel" << endl);
    
    world = new Worldmodel (numSnapshots, comm);

    writeSnaps = false;
    if (logging_snap) {
      writeSnaps = true;
      // open snap_log file
      ostrstream tmp;
      tmp << SNAP_LOG_FILENAME << getMyTeam() << getMyNumber() 
	  << SNAP_LOG_EXTENSION;
      char* filename = tmp.str();
      out.open(filename, ios::out);
      // Version control
      out << SNAPLOG_PROTOCOL << endl;
      delete filename;
    }
 
    break;
  case (COMM_ERROR): 
    cerr << "Onlinecoach: unspecified error when initializing !" 
	 << endl; break;
  case (COMM_ERROR_SERVER_NOT_FOUND):
    cerr << "Onlinecoach: could not find server's host !"<< endl; break;
  case (COMM_ERROR_SOCKET_INIT):
    cerr << "Onlinecoach: could not create and bind socket !" << endl; break;
  case (COMM_ERROR_SERVER_INIT):
    cerr << "Onlinecoach: could not deliver checkin message to server !" 
	 << endl; break;
  case (COMM_ERROR_PARSE):
    cerr << "Onlinecoach: could not parse init response !" << endl; break;
  case (COMM_ERROR_NO_DATA):
    cerr << "Onlinecoach: did not receive response from server !" 
	 << endl; break;
  }
       
}


Onlinecoach::~Onlinecoach () {

  TRACE (cout << "Onlinecoach: entering destructor, closing down" << endl);

  delete (world);
  delete (comm);

  if (writeSnaps) out.close();

  TRACE (cout << "Onlinecoach: leaving destructor" << endl);

}


void 
Onlinecoach::coach() {

  if (comm->isServerAlive()) {
    TRACE (cout << "Onlinecoach: started" << endl);

    while (comm->isServerAlive() && !comm->isGameOver()) {
      
      TRACE (cout << "Onlinecoach: waiting for signals and handling them" 
	     << endl);
      comm->drainSocket();

      if (comm->isActionNeeded()) {
	
	TRACE (cout << "Onlinecoach: adjusting worldmodel" << endl);
	world->adjust();
	
	TRACE (cout << "Onlinecoach: getting snapshot of time " 
	       << getTime() << endl);
	
	Snapshot* snap = world->getSnapshot(getTime());
	
	if (snap == NULL) {
	  cerr << "Onlinecoach (" << getMyTeam()
	       <<"): cannot generate Snapshot at time " << getTime() << endl;
	} else {
	  
	  if ( writeSnaps && comm->isTimeRunning() ) {
	    snap->binaryWriteToStreamOSI(out);
	    out << endl;
	  }
	  
	  TRACE (cout << "Onlinecoach: analyzing situation" << endl);
	  analyze(snap);
	  
	  if (getPlayMode() != PLAY_MODE_PLAY_ON) {
	    TRACE (cout << "Onlinecoach: play_mode != play_on, acting" 
		   << endl);
	    act(snap);
	  }
	
	}
	
      } else {
	TRACE(cout << "Onlinecoach: no new timestep yet" << endl);
      }

    } // while server alive

    TRACE (cout << "Onlinecoach: server is gone, leaving" << endl);
    
  } else { 
    // server never was alive
    cerr << "Onlinecoach: could not connect to sserver, closing down" << endl;
  }
     
} //play
