//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player class  GoalShotEval
//
// \$Id: goalshoteval.C,v 1.21 2002/06/06 15:59:23 fel Exp $
// \$manager: $schappel
//
//////////////////////////////////////////////////////////////////////

#include "goalshoteval.h"


// the trace macros
////////////////////////

#undef USEGOALEVALTRACE
#undef USEGOALDRAWTRACE

// trace macro: uncomment the line you want

// to see the players decision process in FUNSSEL log window
//#define USEGOALEVALTRACE

// to see some graphical things in FUNSSEL
//#define USEGOALDRAWTRACE

#ifdef USEGOALEVALTRACE
#define GOALEVALTRACE(cmd) cmd 
#warning GoalShotEval debug trace GOALEVALTRACE is activated!
#else
#define GOALEVALTRACE(cmd)
#endif

#ifdef USEGOALDRAWTRACE
#define GOALDRAWTRACE(cmd) cmd
#warning GoalShotEval debug trace GOALDRAWTRACE is activated!
#else
#define GOALDRAWTRACE(cmd)
#endif


////////////////////////////
//
// Constructor
//
///////////////////////////

GoalShotEval::GoalShotEval()
  // quasi-constants, perhaps with an adaptive behaviour?  
  // estimated values!?

{
  // the opening angle of the corridor from the player to the selected
  // target
  // [C] eventuell ins agent.conf uebernehmen
  corridorAngle = PI_6;

  // reading constants from agent.conf

  // minimum and maximum distances to the edges
  minDistToCorner = Config::getValueAsDouble("goal_minDistToCorner");
  maxDistToCorner = Config::getValueAsDouble("goal_maxDistToCorner");

  // important targetpositions for our team in the others goal
  leftEdgeOfGoal  = Vector2d(FIELD_LENGTH /2., -GOAL_WIDTH / 2.);
  rightEdgeOfGoal = Vector2d(FIELD_LENGTH /2.,  GOAL_WIDTH / 2.);
  middleOfGoal    = Vector2d(FIELD_LENGTH /2., 0);

  // important targetpositions for the other team in our goal
  leftEdgeOfOurGoal  = Vector2d(-FIELD_LENGTH /2., -GOAL_WIDTH / 2.);
  rightEdgeOfOurGoal = Vector2d(-FIELD_LENGTH /2.,  GOAL_WIDTH / 2.);
  middleOfOurGoal    = Vector2d(-FIELD_LENGTH /2., 0 );


  // elder things will not be considered
  maxAge = Config::getValueAsInt("goal_maxAge");

  maxKickDuration = Config::getValueAsInt("goal_maxKickDuration");       
  
  maxKickLength  = Config::getValueAsDouble("goal_maxKickLength");

  corridorWidth      = Config::getValueAsDouble("goal_corridorWidth");
  corridorWidthSmall = Config::getValueAsDouble("goal_corridorWidthSmall");

  distGradeIsImpDist    = Config::getValueAsDouble("goal_distGradeIsImpDist");
  distGradeIsBadDist    = Config::getValueAsDouble("goal_distGradeIsBadDist");
  distGradeIsPoorDist   = Config::getValueAsDouble("goal_distGradeIsPoorDist");
  distGradeIsAverageDist = Config::getValueAsDouble("goal_distGradeIsAverageDist");
  distGradeIsGoodDist   = Config::getValueAsDouble("goal_distGradeIsGoodDist");
  anyCaseGoalShotDist   = Config::getValueAsDouble("goal_anyCaseGoalShotDist");
  
  xDistToGoalLine       = Config::getValueAsDouble("goal_xDistToGoalLine");

  goalieNearDist  = Config::getValueAsDouble("goal_goalieNearDist");
  goalLineImpDist = Config::getValueAsDouble("goal_goalLineImpDist");
};

////////////////////////////
//
// Destructor
//
///////////////////////////

GoalShotEval::~GoalShotEval()
{};


/////////////////////////////////////////////////////////////////////////
//
// public methods
//
/////////////////////////////////////////////////////////////////////////


//////////////////////////////////////
//
// Grade GoalShotEval::ourEval
//
/////////////////////////////////////

Grade GoalShotEval::ourEval(const Snapshot* worldSnap, 
			    Vector2d playerpos) const
{
  Grade grade;

  // do we have a valid snapshot?
  if(worldSnap == NULL)
    {
      cerr << "GOAL: ourEval -> snapshot is NULL" ;
      grade.set_IMPOSSIBLE();
      return grade;
    }

  // calculate the minimum of the player's distance to the left edge,
  // the middle or the right edge of the goal for calculating a first
  // goalshot target

  double minGoalDist;
  Vector2d goalTarget;
  Vector2d leftTarget;
  Vector2d rightTarget;
  Vector2d middleTarget;

  targetInGoal( playerpos, goalTarget, leftTarget, middleTarget, 
		rightTarget, minGoalDist );

  // searching for the best corridor and evaluating it
  grade = searchAndEvalOurBestCorridor(worldSnap, 
				       playerpos,
				       goalTarget,
				       leftTarget,
				       middleTarget,
				       rightTarget);


  // if the goalie is to catch the ball in the next step, we shoot anyway
  if ( (worldSnap->otherGoalieNo>0) &&  (worldSnap->otherGoalieNo<12) &&
       (worldSnap->otherTeam(worldSnap->otherGoalieNo).distTo(worldSnap->ball)
      < CATCHABLE_AREA_L + PLAYER_SPEED_MAX) ){ 
      grade.set_PERFECT();
      GOALEVALTRACE( { ostrstream ost;
      ost << "GOAL: goalie near, shooting anyway" << char(0);
      
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
  }
  

  return grade;
};


////////////////////////////////////
//
// Grade GoalShotEval::otherEval
//
///////////////////////////////////

Grade GoalShotEval::otherEval(const Snapshot* worldSnap, 
			      Vector2d playerpos) const
{
  Grade grade;

  // do we have a valid snapshot?
  if(worldSnap == NULL)
    {
      cerr << "GOAL: otherEval(" << playerpos << ") -> snapshot is NULL" ;
      grade.set_IMPOSSIBLE();
      return grade;
    }

  // calculate the minimum of the player's distances to the left edge,
  // the middle or the right edge of the goal for calculating a first
  // goalshot target

  double minGoalDist;
  Vector2d goalTarget;
  Vector2d leftTarget;
  Vector2d middleTarget;
  Vector2d rightTarget;

  targetInGoal( playerpos, goalTarget, leftTarget, middleTarget, 
		rightTarget, minGoalDist );

  // searching for the best corridor and evaluating it
  grade = searchAndEvalOtherBestCorridor( worldSnap, playerpos, 
					  goalTarget,
					  leftTarget, 
					  middleTarget, 
					  rightTarget);
  return grade;
};  


//////////////////////////////////////////////////////////////////////////
//
// private methods
//
//////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////
//
// GoalShotEval::targetInGoal
//
//////////////////////////////////////////

void GoalShotEval::targetInGoal(Vector2d playerpos, 
				Vector2d &minTarget,
				Vector2d &leftTarget,
				Vector2d &middleTarget, 
				Vector2d &rightTarget,
				double &minDist) const
{
  // correct the positions in the goal
  leftTarget = leftEdgeOfGoal + correctCorner( dist( playerpos, leftEdgeOfGoal ) );
  middleTarget = middleOfGoal;
  rightTarget = rightEdgeOfGoal - correctCorner( dist( playerpos, rightEdgeOfGoal ) );

  GOALEVALTRACE( { ostrstream ost;
  ost << "GOAL: targetInGoal(" << playerpos 
      << ") -> new targets are "
      << leftTarget << " " << middleTarget << " "
      << rightTarget << char(0);

  char * str = ost.str(); 
  Comm::sendMCmd(MLog(str));
  delete str;
  });
 
  // calculate the distances to the positions in the goal
  double distToLeft  = dist( playerpos, leftTarget );
  double distToMid   = dist( playerpos, middleTarget );
  double distToRight = dist( playerpos, rightTarget );

  minDist = 0;

  if( distToLeft < distToMid)
    if(distToRight < distToLeft)
      {
	minDist = distToRight;
	minTarget = rightTarget;
      }
    else
      {
	minDist = distToLeft;
	minTarget = leftTarget;
      }
  else
    if(distToRight < distToMid)
      {
	minDist = distToRight;
	minTarget = rightTarget;
      }
    else
      {
	minDist = distToMid;
	minTarget = middleTarget;
      }

  GOALEVALTRACE( { ostrstream ost;
  ost << "GOAL: targetInGoal(" << playerpos 
      << ") -> returns target = "
      << minTarget << ", minDist = "
      << minDist << char(0);
  char * str = ost.str(); 
  Comm::sendMCmd(MLog(str));
  delete str;
  });
}


///////////////////////////////////////////
//
// GoalShotEval::correctCorner
//
//////////////////////////////////////////

Vector2d GoalShotEval::correctCorner(double dist) const
{
  if ( dist > distGradeIsImpDist )
    return Vector2d( 0.0, maxDistToCorner );

  if ( dist > anyCaseGoalShotDist )
    return Vector2d( 0.0, minDistToCorner + ( maxDistToCorner - minDistToCorner ) *
		     ( dist               - anyCaseGoalShotDist ) /
		     ( distGradeIsImpDist - anyCaseGoalShotDist ) );

  return Vector2d( 0.0, minDistToCorner );
}


///////////////////////////////////////////
//
// GoalShotEval::availableTimeForKick
//
//////////////////////////////////////////

int GoalShotEval::availableTimeForKick(const Snapshot *worldSnap, 
				       Vector2d playerpos, 
				       team_type team) const
{
  if(worldSnap == NULL)
    {
      cerr << "GOAL: availableTimeForKick -> invalid worldSnap" ;
      return 0;
    }

  // identify the closest opponent

  int closestOppNo;
  if(team == OUR_TEAM)
    // identify the member of the other team being next to the player with 
    // no "player" of our team
    closestOppNo = worldSnap->otherTeamClosestTo(playerpos, maxAge);
  else
    // identify the member of our team being next to the player with no 
    // "player" of the other team
    closestOppNo = worldSnap->ourTeamClosestTo(playerpos, maxAge);
    
  // player sees no opponent, so everything is allright
  if (closestOppNo == UNKNOWN) 
    return maxKickDuration;               

  // calculating the distance to the nearest player

  double distToOppPlayer;
  int availableTime = 0;
  
  if (team == OUR_TEAM)
    {
      distToOppPlayer 
	= dist(worldSnap->otherTeam(closestOppNo).pos, playerpos);
      availableTime = int(distToOppPlayer / worldSnap->otherTeam(closestOppNo).vmax);
    }
  else
    {
      distToOppPlayer
	= dist(worldSnap->ourTeam(closestOppNo).pos, playerpos);
      availableTime = int(distToOppPlayer / worldSnap->ourTeam(closestOppNo).vmax);
    }

  // if there might have been a ball collision don`t place
  if( worldSnap->collState != NONE )
    availableTime = 1;

  return min( maxKickDuration, max( availableTime, 1 ) );
}


///////////////////////////////////////////////
//
// GoalShotEval::distToLine
//
//////////////////////////////////////////////

double GoalShotEval::distToLine(Vector2d object, Vector2d source, 
				Vector2d target) const
{
  double sx = source.getX();
  double sy = source.getY();

  double tx = target.getX();
  double ty = target.getY();

  double ox = object.getX();
  double oy = object.getY();

  // project object onto the line source->target
  double proj = ( ( (sx-tx) * (sx-ox) + (sy-ty) * (sy-oy) ) /
		  ( (sx-tx) * (sx-tx) + (sy-ty) * (sy-ty) ) );
  
  if ( proj < 0 )
    // projection point is in front of the source
    return dist(object, source);
  
  if( proj > 1)
    return dist(object, target); 

  // calculate distance between object and the line source->target
  return sqrt( sqr( ox - sx - proj * (tx-sx) ) 
	       + sqr( oy - sy - proj * (ty-sy) ) );

}


///////////////////////////////////////////
//
// GoalShotEval::evaluateCorridor
//
//////////////////////////////////////////

Grade GoalShotEval::evaluateCorridor(const Snapshot *worldSnap, 
				     Vector2d playerpos,  
				     Vector2d target, Angle angle,
				     Grade startGrade,
				     team_type team) const
{
  Grade grade, distgrade, corridorgrade;

  // calculating the grade for the dist between the player and target
  // if the distance from the player to the target
  //       is greater than distGradeIsImpDist      : distgrade = IMPOSSIBLE
  //       is greater than distGradeIsBadDist
  //       and smaller than distGradeIsImpDist     : distgrade = BAD
  //       is greater than distGradeIsPoorDist
  //       and smaller than distGradeIsBadDist     : distgrade = POOR
  //       is greater than distGradeIsAverageDist
  //       and smaller than distGradeIsPoorDist    : distgrade = AVERAGE
  //       is greater than distGradeIsGoodDist
  //       and smaller than distGradeIsAverageDist : distgrade = GOOD
  //       is smaller than distGradeIsGoodDist     : distgrade = PERFECT
  // if the distance from the player to the target is even smaller than 
  // anyCaseGoalShotDist then PERFECT is returned
  
  double distToTarget = dist(playerpos, target);

  // [!] HACK : use distance to far corner
  distToTarget = max( dist( playerpos, leftEdgeOfGoal ),
		      max( dist( playerpos, middleOfGoal ),
			   dist( playerpos, rightEdgeOfGoal ) ) );

  if ( distToTarget <= anyCaseGoalShotDist )
    // the distance to the target is very small, so try to kick the ball 
    // in any case
    
    // [!] HACK : only if the goalie could not reach the ball
    if (!worldSnap->otherTeamInCorridor(worldSnap->ball.pos,target,PI/2,5))
    {
      grade.set_PERFECT();
      GOALEVALTRACE( { ostrstream ost;
      ost << "GOAL: evaluateCorr(" << playerpos 
	  << ") ->distance to target " << target
	  << " is " << distToTarget 
	  << ", grade = " << grade
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
      return grade;
    }

  if( distToTarget > distGradeIsImpDist ) {
    // This is near to technically impossible (in a normal game), so set
    // grade to impossible and return
    distgrade.set_IMPOSSIBLE();

    // trace and return are for speed reasons because our agents
    // often have timing problems
    GOALEVALTRACE( { ostrstream ost;
    ost << "GOAL: evaluateCorr(" << playerpos 
	<< ") -> distance of player to target " 
	<< target << " is " 
	<< distToTarget << ", distgrade = "
	<< distgrade
	<< char(0);
    char * str = ost.str(); 
    Comm::sendMCmd(MLog(str));
    delete str;
    });

    return distgrade;
  } 
  else if( distToTarget > distGradeIsBadDist )
    distgrade.set_BAD();
  else if( distToTarget > distGradeIsPoorDist )
    distgrade.set_POOR();
  else if( distToTarget > distGradeIsAverageDist) 
    distgrade.set_AVERAGE();
  else if( distToTarget > distGradeIsGoodDist )
    distgrade.set_GOOD();
  else
    distgrade.set_PERFECT();
  
  GOALEVALTRACE( { ostrstream ost;
  ost << "GOAL: evaluateCorr(" << playerpos 
      << ") -> distance of player to target " 
      << target << " is " 
      << distToTarget << ", distgrade = "
      << distgrade
      << char(0) ;
  char * str = ost.str(); 
  Comm::sendMCmd(MLog(str));
  delete str;
  });

  // if the angle between the goalline and the player becomes
  // too small, the player has to be closer to the goal
  // use fabs for x coordinate so that estimate works for both teams (CM)
  if ( ( fabs(playerpos.getX()) > FIELD_LENGTH / 2. - xDistToGoalLine )
       && ( fabs(playerpos.getY()) > PENALTY_AREA_WIDTH /2. ))
    {
      distgrade.downgrade();
      GOALEVALTRACE( { ostrstream ost;
      ost << "GOAL: evaluateCorr(" << playerpos 
	  << ") -> angle to goal line too small, downgrade distgrade to "
	  << distgrade
	  << char(0) ;
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
    }
  
  // calculating the grade for the corridor itself and then taking the 
  // lower of this grade and the distgrade

  corridorgrade = startGrade;

  int numberOfOpponents;
  
  if(team == OUR_TEAM)
    numberOfOpponents 
      = worldSnap->otherTeamInCorridor(playerpos, 
				       target, corridorWidth, 
				       angle, maxAge);
  else
    numberOfOpponents 
      = worldSnap->ourTeamInCorridor(playerpos, 
				     target, corridorWidth, 
				     angle, maxAge);
    
  if(numberOfOpponents == 0)
    {
      // corridorgrade stays at the startvalue
      GOALEVALTRACE({ ostrstream ost;
      ost << "GOAL: evaluateCorr(" << playerpos 
	  << ") -> corridor towards position " 
	  << target << " with width " 
	  << corridorWidth << " is free; corridorgrade = "
	  << corridorgrade
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
		    
    }
  else
    // corridorgrade will be downgraded
    {
      corridorgrade.downgrade();
      // there are opponents in the corridor, so let us look if there is
      // no opponent whose kickrange will be crossed by the ball 
      // (supposed none of the players is moving) 
      // [C] haengt jetzt von der Art des Gegeners ab, worst case-Betrachtung??
      if(team == OUR_TEAM)
	numberOfOpponents =
	  worldSnap->otherTeamInCorridor(playerpos, 
					 target, corridorWidthSmall, 
					 angle, maxAge);
      else
	numberOfOpponents =
	  worldSnap->ourTeamInCorridor(playerpos, 
				       target, corridorWidthSmall, 
				       angle, maxAge);
      
      if(numberOfOpponents == 0)
	{
	  GOALEVALTRACE({ ostrstream ost;
	  ost << "GOAL: evaluateCorr (" << playerpos 
	      << ") -> corridor towards position " 
	      << target <<" with width "
	      << corridorWidthSmall << " is free; corridorgrade = "
	      << corridorgrade 
	      << char(0);
	  char * str = ost.str(); 
	  Comm::sendMCmd(MLog(str));
	  delete str;
	  });
	}
      else
	// corridorgrade will be downgraded again
	{
	  corridorgrade.downgrade();
	  // there are even opponents in the corridor with width 
	  // corridorWidthSmall
	  // let us look if the corridor with width 0.5 * corridorWidthSmall
	  // is free
	  // [C] was mit dem Faktor 0.5 machen???
	  if(team == OUR_TEAM)
	    numberOfOpponents = 
	      worldSnap->otherTeamInCorridor(playerpos, 
					     target, 0.5*corridorWidthSmall, 
					     angle, maxAge);
	  else
	    numberOfOpponents = 
	      worldSnap->ourTeamInCorridor(playerpos, 
					   target, 0.5*corridorWidthSmall, 
					   angle, maxAge);
	    
	  if(numberOfOpponents == 0)
	    {
	      GOALEVALTRACE( { ostrstream ost;
	      ost << "GOAL: evaluateCorr (" << playerpos 
		  << ") -> corridor towards position " 
		  << target <<" with width "
		  << 0.5*corridorWidthSmall 
		  << " is free; corridorgrade = "
		  << corridorgrade
		  << char(0);
	      char * str = ost.str(); 
	      Comm::sendMCmd(MLog(str));
	      delete str;
	      });
	    }
	  else
	    // corridorgrade will be downgraded for the third time
	    {
	      corridorgrade.downgrade();
	      GOALEVALTRACE({ostrstream ost;
	      ost << "GOAL: evaluateCorr (" << playerpos 
		  << ") -> even corridor towards position " 
		  << target <<" with smallest width "
		  << 0.5*corridorWidthSmall 
		  << " is not free; corridorgrade = "
		  << corridorgrade
		  << char(0);
	      char * str = ost.str(); 
	      Comm::sendMCmd(MLog(str));
	      delete str;
	      });
	    }
	}
    }
  
  // calculating the worse of distgrade and corridorgrade
  if (distgrade.is_worse_than(corridorgrade))
    grade = distgrade;
  else 
    grade = corridorgrade;

  GOALEVALTRACE( { ostrstream ost;
  ost << "GOAL: evaluateCorr (" << playerpos 
      << ") -> worse of corridorgrade and distgrade is " 
      << grade
      << char(0);
  char * str = ost.str(); 
  Comm::sendMCmd(MLog(str));
  delete str;
  });

  return grade;
}


///////////////////////////////////////////////
//
// GoalShotEval::baseSearchAndEvalForCorridor
//
//////////////////////////////////////////////
 
Grade GoalShotEval::baseSearchAndEvalForCorridor(const Snapshot *worldSnap, 
						 Vector2d playerpos,
						 Vector2d &target, 
						 Vector2d leftTarget,
						 Vector2d middleTarget,
						 Vector2d rightTarget,
						 bool useLeft, 
						 bool useMiddle,
						 bool useRight) const
{
  Grade grade;
  grade.set_PERFECT();

  // if all targets are too far away, don't try
  if( !useLeft && !useMiddle && !useRight )
    {
      grade.set_IMPOSSIBLE();
      return grade;
    }

  // security distances to the border:
  // [C] abhngig von der Spielergroesse!!!!!
  // has to be changed for heterogenous players!
  const double SECURITY_DIST = KICKABLE_MARGIN + BALL_SIZE + PLAYER_SIZE;   // = 1.085 m

  // If we are near to the opponent's goal line, check if
  // we are not near the goal
  ////////////////////////////////////////
  if( ( playerpos.getX() > FIELD_LENGTH/2.0 - SECURITY_DIST ) &&
      ( fabs( playerpos.getY() ) > GOAL_WIDTH/2.0 + goalLineImpDist ) ) 
    {
      grade.set_IMPOSSIBLE();
      GOALEVALTRACE ({ostrstream ost;
      ost << "GOAL: baseSearch (" << playerpos
	  << ") -> too near to goal line & too far away from goal; grade = " 
	  << grade 
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
      return grade;
    }
  
  // draw corridors
  GOALDRAWTRACE(Comm::sendMCmd(MDCorridor(playerpos, leftTarget, corridorWidth, 
					  corridorAngle)));
  GOALDRAWTRACE(Comm::sendMCmd(MDCorridor(playerpos, middleTarget, corridorWidth, 
					  corridorAngle)));
  GOALDRAWTRACE(Comm::sendMCmd(MDCorridor(playerpos, rightTarget, corridorWidth, 
					  corridorAngle)));


  // in which of the corridors does the goalie stand ? 
  // corridors with goalie in them should not be used

  bool goalieKnown = ( ( worldSnap->otherGoalieNo != UNKNOWN ) && 
		       ( worldSnap->otherTeam( worldSnap->otherGoalieNo ).age < maxAge ) );

  if( goalieKnown )
    {
      const WorldPlayer &goalie = worldSnap->otherTeam( worldSnap->otherGoalieNo );

      useLeft   = useLeft   && !goalie.inCorridor( playerpos, leftTarget, corridorWidth, corridorAngle );
      useMiddle = useMiddle && !goalie.inCorridor( playerpos, middleTarget, corridorWidth, corridorAngle );
      useRight  = useRight  && !goalie.inCorridor( playerpos, rightTarget, corridorWidth, corridorAngle );

      if( !( useLeft && useMiddle && useRight ) )
	{
	  GOALEVALTRACE({ostrstream ost;
	  ost << "GOAL: baseSearch (" << playerpos 
	      << ") -> goalie in " << ( !useLeft ? "[left]" : "" )
	      << ( !useMiddle ? "[middle]" : "" ) << ( !useRight ? "[right]" : "" )
	      << " corridor" << char(0);
	  char * str = ost.str(); 
	  Comm::sendMCmd(MLog(str));
	  delete str;
	  });
	}
    }

  // decide which corridor should be used

  // special case: all the corridors are unusable, that means the player is too far 
  // away from the goal and / or the goalie is standing in all the corridors
  if( !useLeft && !useMiddle && !useRight )
    {
      Vector2d mirrorTarget = target;
      mirrorTarget.setY( -target.getY() );

      if( dist( playerpos, mirrorTarget ) <= anyCaseGoalShotDist )
	// if( dist( playerpos, target ) < anyCaseGoalShotDist )
 	// the dist to the target is very short, so try it in any case
 	{
 	  grade.set_PERFECT();
 	  GOALEVALTRACE({ostrstream ost;
	  ost << "GOAL: baseSearch (" << playerpos 
	      << ") -> dist to target "
	      << target << " smaller than " << anyCaseGoalShotDist
	      << "; grade = " << grade 
	      << char(0);
	  char * str = ost.str(); 
	  Comm::sendMCmd(MLog(str));
	  delete str;
	  });
 	} // dist to target small 
      else 
 	{  
	  // if next to goalie, try anyway ...
	  if( goalieKnown && 
	      dist( playerpos, worldSnap->otherTeam( worldSnap->otherGoalieNo ).pos ) < goalieNearDist )
	    {
	      grade = evaluateCorridor( worldSnap, playerpos, target, corridorAngle, grade );
	      GOALEVALTRACE({ostrstream ost;
	      ost << "GOAL: baseSearch (" << playerpos 
		  << ") -> all corridors unusable, but goalie near,"
		  << " kick to " << target << " grade = " << grade
		  << char(0);
	      char * str = ost.str(); 
	      Comm::sendMCmd(MLog(str));
	      delete str;
	      });
	    } 
	  else 
	    {
	      grade.set_BAD();
	      GOALEVALTRACE({ostrstream ost;
	      ost << "GOAL: baseSearch (" << playerpos 
		  << ") -> all corridors unusable; grade = "
		  << grade 
		  << char(0);
	      char * str = ost.str(); 
	      Comm::sendMCmd(MLog(str));
	      delete str;
	      });
	    }
 	} // dist to target not small
      return grade;
    } // all corridors unusable

  // here at least one corridor is usable
  if( goalieKnown )
    {
      // ---< case 1 >--- : goalie is known
      // in this case, choose target line that goalie reaches latest

      const WorldPlayer &goalie = worldSnap->otherTeam( worldSnap->otherGoalieNo );

      GOALEVALTRACE({ ostrstream ost;
      ost << "GOAL: baseSearch (" << playerpos 
	  << ") -> goalie is known; usable corridors: " << ( useLeft ? "left " : "" )
	  << ( useMiddle ? "middle " : "" ) << ( useRight ? "right " : "" )
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });

      double distToLeft   = useLeft   ? distToLine( goalie.pos, playerpos, leftTarget )   : 0.0;
      double distToMiddle = useMiddle ? distToLine( goalie.pos, playerpos, middleTarget ) : 0.0;
      double distToRight  = useRight  ? distToLine( goalie.pos, playerpos, rightTarget )  : 0.0;

      useLeft = useLeft && ( distToLeft >= distToMiddle ) && ( distToLeft >= distToRight );
      if( useLeft )
	target = leftTarget;
      else
	{
	  useMiddle = useMiddle && ( distToMiddle >= distToRight );
	  if( useMiddle )
	    target = middleTarget;
	  else
	    target = rightTarget;
	}
    }
  else
    {
      // ---< case 2 >--- : goalie is unknown (should only happen in the first few cycles)
      // in this case, choose target whose corridor is emptiest

      GOALEVALTRACE({ ostrstream ost;
      ost << "GOAL: baseSearch (" << playerpos 
	  << ") -> goalie is Unknown; usable corridors: " << ( useLeft ? "[left]" : "" )
	  << ( useMiddle ? "[middle]" : "" ) << ( useRight ? "[right]" : "" )
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });

      int oppInLeftCorr
	= useLeft   ? worldSnap->otherTeamInCorridor( playerpos, leftTarget, 
						      corridorWidth, corridorAngle, maxAge ) : 12;
      int oppInMiddleCorr
	= useMiddle ? worldSnap->otherTeamInCorridor( playerpos, middleTarget, 
						      corridorWidth, corridorAngle, maxAge ) : 12;
      int oppInRightCorr
	= useRight  ? worldSnap->otherTeamInCorridor( playerpos, rightTarget, 
						      corridorWidth, corridorAngle, maxAge ) : 12;

      useLeft = useLeft && ( oppInLeftCorr < oppInMiddleCorr ) && ( oppInLeftCorr < oppInRightCorr );
      if( useLeft )
	target = leftTarget;
      else
	{
	  useMiddle = useMiddle && ( oppInMiddleCorr <= oppInRightCorr );
	  if( useMiddle )
	    target = middleTarget;
	  else
	    target = rightTarget;
	}
    }

  GOALEVALTRACE({ ostrstream ost;
  ost << "GOAL: baseSearch (" << playerpos 
      << ") -> using " << ( useLeft ? "left" : useMiddle ? "middle" : "right" )
      << " corridor" << char(0);
  char * str = ost.str(); 
  Comm::sendMCmd(MLog(str));
  delete str;
  });
  
  // corridor is chosen, now evaluate it
  grade = evaluateCorridor( worldSnap, playerpos, target, corridorAngle, grade );
  return grade;
}


///////////////////////////////////////////////
//
// GoalShotEval::searchAndOurEvalBestCorridor
//
//////////////////////////////////////////////

Grade GoalShotEval::searchAndEvalOurBestCorridor(const Snapshot *worldSnap, 
						 Vector2d playerpos,
						 Vector2d &target,
						 Vector2d leftTarget,
						 Vector2d middleTarget,
						 Vector2d rightTarget) const
{
  Grade grade;

  // now the player looks for the best corridor towards the goal
  // by first deciding which corridors are possible by the given maxKickLength

  bool useLeftCorridor = true;
  bool useMiddleCorridor = true;
  bool useRightCorridor = true;
  
  if( dist( playerpos, leftTarget ) > maxKickLength )
    {
      GOALEVALTRACE({ ostrstream ost;
      ost << "GOAL: bestCorr(" << playerpos 
	  << ") -> left Edge too far away; left unusable"
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
      useLeftCorridor = false;
    }

  if( dist( playerpos, middleTarget ) > maxKickLength )
    {
      GOALEVALTRACE({ ostrstream ost;
      ost << "GOAL: bestCorr(" << playerpos 
	  << ") -> middle of goal too far away; middle unusable"
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
      useMiddleCorridor = false;
    }

  if( dist( playerpos, rightTarget ) > maxKickLength )
    {
      GOALEVALTRACE({ ostrstream ost;
      ost << "GOAL: bestCorr(" << playerpos 
	  << " -> right Edge too far away; right unusable"
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
      useRightCorridor = false;
    }

  grade = baseSearchAndEvalForCorridor(worldSnap, playerpos, target,
				       leftTarget, middleTarget, rightTarget,
				       useLeftCorridor, useMiddleCorridor, 
				       useRightCorridor);
  return grade;
}


////////////////////////////////////////////////////
//
// GoalShotEval::searchAndEvalOtherBestCorridor
//
///////////////////////////////////////////////////


Grade GoalShotEval::searchAndEvalOtherBestCorridor(const Snapshot *worldSnap, 
						   Vector2d playerpos,
						   Vector2d &target,
						   Vector2d leftTarget,
						   Vector2d middleTarget,
						   Vector2d rightTarget) const
{
  Grade grade;
  grade.set_PERFECT();

  // now the player looks for the best corridor towards the goal
  // by first deciding which corridors(s) are possible by the given 
  // maxKickLength

  bool useLeft   = true;
  bool useMiddle = true;
  bool useRight  = true;
  
  if( dist( playerpos, leftTarget ) > maxKickLength )
    {
      GOALEVALTRACE({ostrstream ost;
      ost << "GOAL: otherBestCorr" << playerpos 
	  << " -> left edge too far away; left unusable"
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
      useLeft = false;
    }
  if( dist( playerpos, middleTarget ) > maxKickLength )
    {
      GOALEVALTRACE({ostrstream ost;
      ost << "GOAL: otherBestCorr" << playerpos 
	  << " -> middle of goal too far away; middle unusable"
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
      useMiddle = false;
    }
  if( dist( playerpos, rightTarget ) > maxKickLength )
    {
      GOALEVALTRACE({ostrstream ost;
      ost << "GOAL: otherBestcorr" << playerpos 
	  << " -> right edge too far away; right unusable"
	  << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
      useRight = false;
    }

  if( !useLeft && !useMiddle && !useRight )
    {
      grade.set_BAD();
      return grade;
    }

  // in which of the corridors does the goalie stand ? 
  // corridors with goalie in them should not be used

  if( ( worldSnap->ourGoalieNo != UNKNOWN ) && 
      ( worldSnap->ourTeam( worldSnap->ourGoalieNo ).age < maxAge ) )
    {
      const WorldPlayer &goalie = worldSnap->ourTeam( worldSnap->ourGoalieNo );

      if( goalie.inCorridor( playerpos, leftTarget, corridorWidth, corridorAngle ) )
	{
	  GOALEVALTRACE({ostrstream ost;
	  ost << "GOAL: otherBestCorr" << playerpos 
	      << " -> goalie in left corridor"
	      << char(0);
	  char * str = ost.str(); 
	  Comm::sendMCmd(MLog(str));
	  delete str;
	  });
	  useLeft = false;
	}

      if( goalie.inCorridor( playerpos, middleTarget, corridorWidth, corridorAngle ) )
	{
	  GOALEVALTRACE({ostrstream ost;
	  ost << "GOAL: otherBestCorr" << playerpos 
	      << "-> goalie in middle corridor"
	      << char(0);
	  char * str = ost.str(); 
	  Comm::sendMCmd(MLog(str));
	  delete str;
	  });
	  useMiddle = false;
	}

      if( goalie.inCorridor( playerpos, rightTarget, corridorWidth, corridorAngle ) )
	{
	  GOALEVALTRACE({ostrstream ost;
	  ost << "GOAL: otherBestCorr" << playerpos 
	      << " -> goalie in right corridor"
	      << char(0);
	  char * str = ost.str(); 
	  Comm::sendMCmd(MLog(str));
	  delete str;
	  });
	  useRight = false;
	}
    }

  // deciding which corridor should be used

  // all the corridors are unusable, that means the player is too far 
  // away from the goal and / or the goalie stands in the corridor
  if( !useLeft && !useMiddle && !useRight )
    {
      grade.set_POOR();
      GOALEVALTRACE({ostrstream ost;
      ost << "GOAL: otherBestCorr" << playerpos
	  << " -> all corridors unusable; grade = "
	  << grade << char(0);
      char * str = ost.str(); 
      Comm::sendMCmd(MLog(str));
      delete str;
      });
		    
      return grade;
    }

  // for all other cases we need the number of opponents in each of the corridors

  // count the opponent players in the left, right and middle corridor from 
  // the players position to the left /right edge /the middle of the goal
  // corridor has angle corridorAngle

  GOALEVALTRACE({ ostrstream ost;
  ost << "GOAL: otherBestCorr (" << playerpos 
      << ") -> usable corridors: " << ( useLeft ? "[left]" : "" )
      << ( useMiddle ? "[middle]" : "" ) << ( useRight ? "[right]" : "" )
      << char(0);
  char * str = ost.str(); 
  Comm::sendMCmd(MLog(str));
  delete str;
  });

  int oppInLeftCorr
    = useLeft   ? worldSnap->ourTeamInCorridor( playerpos, leftTarget, 
						corridorWidth, corridorAngle, maxAge ) : 12;
  int oppInMiddleCorr
    = useMiddle ? worldSnap->ourTeamInCorridor( playerpos, middleTarget, 
						corridorWidth, corridorAngle, maxAge ) : 12;
  int oppInRightCorr
    = useRight  ? worldSnap->ourTeamInCorridor( playerpos, rightTarget, 
						corridorWidth, corridorAngle, maxAge ) : 12;

  useLeft = useLeft && ( oppInLeftCorr < oppInMiddleCorr ) && ( oppInLeftCorr < oppInRightCorr );
  if( useLeft )
    target = leftTarget;
  else
    {
      useMiddle = useMiddle && ( oppInMiddleCorr <= oppInRightCorr );
      if( useMiddle )
	target = middleTarget;
      else
	target = rightTarget;
    }
  
  GOALEVALTRACE({ ostrstream ost;
  ost << "GOAL: otherBestCorr (" << playerpos 
      << ") -> using " << ( useLeft ? "left" : useMiddle ? "middle" : "right" )
      << " corridor" << char(0);
  char * str = ost.str(); 
  Comm::sendMCmd(MLog(str));
  delete str;
  });
  
  // corridor is chosen, now evaluate it
  grade = evaluateCorridor( worldSnap, playerpos, target, corridorAngle, grade, OTHER_TEAM );
  return grade;
}















