//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player PositionEval
//
// \$Id: PositionEval.h,v 1.9 2002/05/21 12:33:06 wachm Exp $
// \$manager: wachm$
//
//////////////////////////////////////////////////////////////////////

#ifndef __POSITIONEVAL_H__   
#define __POSITIONEVAL_H__

#include "grade.h"
#include "snapshot.h"
#include "goalshoteval.h"
#include "passeval.h"
#include "PositionFunctions.h"

// the class PassEval
//////////////////////////

class PositionEval
{
public:
  // constructor
  PositionEval(const GoalShotEval* _goalshoteval, const PassEval* _passeval);
  // destructor
  ~PositionEval();

  Grade ourEval(const Snapshot* worldSnap, int plnumber) const;

  double GetMinX() const;
  double GetMaxX() const;
  //gives the last calculated min. and the max. positions of our players
  //all players should be positioned between MinX and MaxX

  Vector2d GetPlayerPosition(int PlayerNo) const;
  //gives the last calculated  position to which a player is sent by the 
  //positionmodule

  void GetPlayerArea(int PlayerNo, Vector2d &Min,Vector2d &Max) const;
  //gives the area for which the player PlayerNo is responsible

  bool IsInPlayerArea(int PlayerNo, Vector2d Pos) const;
  //returns true if Pos is in the area of PlayerNo

  int GetPlayerWatchNo(int PlayerNo, int WatchNo) const;
  //returns the number of the players to watch

  //these functions are used to actually calculate the positions of
  //our players. Right now they should only be called by the positionmodule.
  //At the moment they are called during the feedbackWorld procedure.
  void CalMinMaxX(const Snapshot *wS) ;
  void CalPlayerPositions(const Snapshot *wS) ;


  double GetOffensive() const;
private:
  // this module is used by the positionmodule of a player, too!
  friend class PositionModule;

  // some things this module needs:
  /////////////////////////////////////////

  const GoalShotEval* goalshoteval;
  const PassEval* passeval;

  /////////////////////////////////////////
  int    time; //time step number of processed worldSnap
  double oldminX; //last x position of our defenders
  double minX; //x positions of our defenders
  double maxX; //x positions of our attackers
  double offensive;//how offensive we are
  double defensive; 

  PlayerPos RealPositions[10]; //positions of our players (playerNo-2)

  PlayerPos CalRealPos(PlayerPos DPos,PlayerPos OPos,Vector2d BallPos); 

  void readAgent_conf();

 
};


#endif
